#!/usr/bin/env python3
""" Main module for insarviz time-series visualization tool.

Create the main window and manage links with other sub-windows and loading
data modules.
Contains the following class:
* MainWindow

and function:
* main
"""

from typing import Optional, Any

import logging
import logging.handlers
import multiprocessing

import warnings

import pathlib

import sys

import datetime

import json

import numpy as np

import rasterio.warp

from PySide6.QtWidgets import (
    QSizePolicy, QApplication, QLabel, QWidget,
    QMainWindow, QFileDialog, QToolBar,
    QDockWidget,
    QVBoxLayout, QHBoxLayout, QSplitter, QMessageBox, QDialog
)

from PySide6.QtGui import (
    QPixmap, QIcon, QKeySequence, QSurfaceFormat, QOpenGLContext, QOffscreenSurface,
    QUndoStack, QAction,  QActionGroup
)

from PySide6.QtCore import Qt, QCoreApplication, Slot, QDir

import PySide6.QtAsyncio as QtAsyncio

import asyncio
from qasync import QEventLoop, QApplication

import pyqtgraph as pg
import pyqtgraph.exporters
from insarviz.exporters.myCSVExporter import myCSVExporter

from insarviz.Loader import Loader
from insarviz.BandSlider import BandSlider
from insarviz.ColormapWidget import ColormapWidget
from insarviz.map.MapModel import MapModel
from insarviz.map.MapView import MapView
from insarviz.map.MinimapView import MinimapView
from insarviz.plot.PlotModel import PlotModel
from insarviz.plot.TemporalPlotView import TemporalPlotWindow
from insarviz.plot.SpatialPlotView import SpatialPlotWindow
import insarviz.version as version

from insarviz.map.layers.RasterLayerDialog import RasterLayerDialog

from insarviz.custom_widgets import IconDockWidget

from insarviz.utils import openUrl, normalize_path

from insarviz.map.layers.Layer import OpenGLLayer, RasterRGBLayer, Raster1BLayer
from insarviz.map.layers.WMTSLayer import WMTSLayer, WMTSLayerDialog
from insarviz.map.layers.XYZLayer import OpenStreetMapLayer, XYZLayer, XYZLayerDialog
from insarviz.map.layers.SwipeTool import SwipeTool
from insarviz.map.layers.LayerView import LayerView
from insarviz.LoggerWidget import LoggerWidget, LogsStore

logging.getLogger("rasterio").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

log_messages = LogsStore()

queue = multiprocessing.Queue()
queue_handler = logging.handlers.QueueHandler(queue)

queue_listener = logging.handlers.QueueListener(queue,
                                                log_messages,
                                                logging.StreamHandler())
queue_listener.start()

class MPFilter:
    # Filter out messages from multiprocessing.Queue, to avoid deadlocks
    # See: https://docs.python.org/3/library/logging.handlers.html#queuehandler
    def filter(self, record):
        return not (record.levelno == logging.DEBUG and record.module == 'multiprocessing')

root_logger = logging.getLogger()
root_logger.addFilter(MPFilter())
root_logger.addHandler(queue_handler)
root_logger.setLevel(logging.DEBUG)

class MainWindow(QMainWindow):
    """
    The main window of insarviz. Contains every widget and every model.
    """

    def __init__(self, config_dict=None):
        """
        :filepath: the file to load
        :config_dict: the configuration dictionary
        """
        super().__init__()
        # various init:
        self.config_dict = config_dict
        self.project_filepath: Optional[pathlib.Path] = None
        self.setWindowTitle("[*]InsarViz")
        self.setMouseTracking(True)
        self.setDockNestingEnabled(True)
        self.undo_stack: QUndoStack = QUndoStack(self)
        self.undo_stack.cleanChanged.connect(self.on_clean_changed)

        # Loader:
        self.loader: Loader = Loader()

        # Models:
        context: QOpenGLContext = QOpenGLContext()
        context.setShareContext(QOpenGLContext.globalShareContext())
        context.create()
        logger.info(f"Open GL version: {context.format().version()}")
        logger.info(f"Open GL renderable type {context.format().renderableType()}")
        if not context.isValid():
            QMessageBox.critical(self, "Insarviz", "OpenGL error: context is not valid")
            raise RuntimeError("Global OpenGL shared context cannot be created")
        offscreen_surface = QOffscreenSurface()
        offscreen_surface.setFormat(context.format())
        offscreen_surface.create()
        OpenGLLayer.context = context
        OpenGLLayer.offscreen_surface = offscreen_surface
        self.map_model = MapModel(self.loader, context, offscreen_surface)
        self.map_model.opened.connect(self.on_map_model_opened)
        self.map_model.closed.connect(self.on_map_model_closed)
        self.plot_model = PlotModel(self.loader)
        self.map_model.layer_model.selection_initialized.connect(self.plot_model.on_selection_init)
        self.map_model.closed.connect(self.plot_model.close)
        self.map_model.closed.connect(self.undo_stack.clear)
        self.map_model.layer_model.add_undo_command.connect(self.undo_stack.push)

        # Layer manager
        self.layer_model = self.map_model.layer_model
        self.layer_widget = LayerView(self.layer_model)
        # new openstreetmap action
        self.new_openstreetmap_action = QAction(
            QIcon('icons:WMS.svg'), "New OpenStreetMap Layer", self)
        self.new_openstreetmap_action.triggered.connect(self.new_openstreetmap_layer)
        # new WMTS action
        self.new_WMTS_action = QAction(QIcon('icons:WMS.svg'), "New WMTS Layer", self)
        self.new_WMTS_action.triggered.connect(self.new_WMTS_layer)
        # new geomap action
        self.new_XYZ_action = QAction(QIcon('icons:WMS.svg'), "New XYZ Layer", self)
        self.new_XYZ_action.triggered.connect(self.new_XYZ_layer)
        # new raster1B action
        self.new_raster1B_action = QAction(QIcon('icons:raster1B.svg'), "New Raster1B Layer", self)
        self.new_raster1B_action.triggered.connect(self.new_raster1B)
        # new rasterRGB action
        self.new_rasterRGB_action = QAction(QIcon('icons:rasterRGB.svg'),
                                            "New RasterRGB Layer", self)
        self.new_rasterRGB_action.triggered.connect(self.new_rasterRGB)
        # new swipe action
        self.new_swipe_action = QAction(QIcon("icons:swipetool.png"), "New Swipe Tool", self)
        self.new_swipe_action.triggered.connect(self.new_swipe_layer)
        # show all action
        self.layer_showall_action = QAction(QIcon('icons:eye_open.svg'), "Show all", self)
        self.layer_showall_action.setToolTip("Show all layers")
        self.layer_showall_action.triggered.connect(self.layer_model.show_all_layers)
        # hide all action
        self.layer_hideall_action = QAction(QIcon('icons:eye_closed.svg'), "Hide all", self)
        self.layer_hideall_action.setToolTip("Hide all layers")
        self.layer_hideall_action.triggered.connect(self.layer_model.hide_all_layers)
        # move up action
        self.layer_moveup_action = QAction(QIcon('icons:arrowup.png'), "Move Up", self)
        self.layer_moveup_action.setToolTip("Move up layer")
        self.layer_moveup_action.triggered.connect(lambda: self.layer_model.move_layer_up(
            self.layer_widget.proxy_model.mapToSource(
                self.layer_widget.selectionModel().currentIndex())))
        self.layer_model.current_movable_up.connect(self.layer_moveup_action.setEnabled)
        # move down action
        self.layer_movedown_action = QAction(QIcon('icons:arrowdown.png'), "Move Down", self)
        self.layer_movedown_action.setToolTip("Move down layer")
        self.layer_movedown_action.triggered.connect(lambda: self.layer_model.move_layer_down(
            self.layer_widget.proxy_model.mapToSource(
                self.layer_widget.selectionModel().currentIndex())))
        self.layer_model.current_movable_down.connect(self.layer_movedown_action.setEnabled)
        # remove action
        self.layer_remove_action = QAction(QIcon('icons:remove.png'), "Remove", self)
        self.layer_remove_action.setToolTip("Remove item")
        self.layer_remove_action.setShortcuts(QKeySequence.StandardKey.Delete)
        self.layer_remove_action.triggered.connect(lambda: self.layer_model.remove(
            self.layer_widget.proxy_model.mapToSource(
                self.layer_widget.selectionModel().currentIndex())))
        self.layer_model.current_removable.connect(self.layer_remove_action.setEnabled)
        # action group
        self.layer_action_group = QActionGroup(self)
        self.layer_action_group.addAction(self.layer_moveup_action)
        self.layer_action_group.addAction(self.layer_movedown_action)
        self.layer_action_group.addAction(self.layer_showall_action)
        self.layer_action_group.addAction(self.layer_hideall_action)
        self.layer_action_group.addAction(self.layer_remove_action)
        self.layer_action_group.addAction(self.new_openstreetmap_action)
        self.layer_action_group.addAction(self.new_WMTS_action)
        self.layer_action_group.addAction(self.new_XYZ_action)
        self.layer_action_group.addAction(self.new_swipe_action)
        self.layer_action_group.addAction(self.new_raster1B_action)
        self.layer_action_group.addAction(self.new_rasterRGB_action)
        self.layer_action_group.setExclusive(False)
        self.layer_action_group.setDisabled(True)
        self.geolocated_action_group = QActionGroup(self)
        self.geolocated_action_group.addAction(self.new_openstreetmap_action)
        self.geolocated_action_group.addAction(self.new_WMTS_action)
        self.geolocated_action_group.addAction(self.new_XYZ_action)
        self.geolocated_action_group.setExclusive(False)
        self.geolocated_action_group.setDisabled(True)
        # layout
        layer_layout = QVBoxLayout()
        layer_toolbar = QToolBar(self)
        layer_toolbar.addAction(self.layer_moveup_action)
        layer_toolbar.addAction(self.layer_movedown_action)
        layer_toolbar.addAction(self.layer_showall_action)
        layer_toolbar.addAction(self.layer_hideall_action)
        layer_toolbar.addAction(self.layer_remove_action)
        layer_layout.addWidget(layer_toolbar)
        layer_layout.addWidget(self.layer_widget)
        layer_manager_widget = QWidget()
        layer_manager_widget.setLayout(layer_layout)
        self.layer_dockwidget = QDockWidget('Layers', self)
        self.layer_dockwidget.setFeatures(QDockWidget.DockWidgetFeature.NoDockWidgetFeatures)
        self.layer_dockwidget.setWidget(layer_manager_widget)

        # Map:
        self.map_widget = MapView(self.map_model, self)
        self.map_widget.setMouseTracking(True)
        self.map_widget.pointer_changed.connect(self.plot_model.update_pointer)
        self.map_widget.interaction_changed.connect(self.on_mapview_interaction_changed)
        self.plot_model.updated_pointer_info.connect(self.map_widget.update_mouse_tooltip)

        # Minimap
        self.minimap_widget = MinimapView(self.map_model, self)
        self.minimap_widget.pan_map_view.connect(self.map_widget.pan)
        self.minimap_widget.zoom_map_view.connect(self.map_widget.zoom)
        self.minimap_widget.set_center_map_view.connect(self.map_widget.set_view_center)
        self.map_widget.viewport_matrix_changed.connect(
            self.minimap_widget.update_mapview_viewport_matrix)
        self.minimap_dock_widget = QDockWidget('Minimap', self)
        self.minimap_dock_widget.setAllowedAreas(Qt.DockWidgetArea.LeftDockWidgetArea)
        self.minimap_dock_widget.setWidget(self.minimap_widget)

        # Colormap
        self.colormap_widget = ColormapWidget(self)
        # connect signals to slots
        self.map_model.closed.connect(self.colormap_widget.on_close)
        self.map_model.total_hist_changed.connect(self.colormap_widget.set_total_histogram)
        self.map_model.band_hist_changed.connect(self.colormap_widget.set_band_histogram)
        self.map_model.request_colormap.connect(self.colormap_widget.set_colormap)
        self.map_model.v0_v1_changed.connect(self.colormap_widget.set_v0_v1)
        self.loader.data_units_loaded.connect(self.colormap_widget.set_data_units)
        self.loader.histograms_computed.connect(self.colormap_widget.on_histograms_computed)
        self.loader.computing_histograms.connect(self.colormap_widget.on_compute_histogram)
        self.colormap_widget.colormap_changed.connect(self.map_model.set_colormap)
        self.colormap_widget.v0_v1_changed.connect(self.map_model.set_v0_v1)
        self.colormap_widget.compute_histograms.connect(self.loader.compute_histograms)
        self.colormap_widget.cancel_compute_histograms.connect(
            self.loader.cancel_compute_histograms)

        # dates slider:
        self.band_slider = BandSlider(self.loader)
        self.band_slider.value_changed.connect(self.map_model.show_band)
        self.map_model.closed.connect(self.band_slider.on_close)
        self.loader.data_loaded.connect(self.band_slider.on_data_loaded)

        # plot windows:
        # temporal window
        self.temporal_plot_window = TemporalPlotWindow(self.plot_model)
        self.band_slider.value_changed.connect(
            self.temporal_plot_window.plot_widget.date_marker.on_slider_changed)
        self.temporal_plot_window.plot_widget.date_marker.pos_changed.connect(
            self.band_slider.set_value)
        self.map_widget.request_lock_axes.connect(lambda: self.temporal_plot_window.checkbox_axes.setCheckState(Qt.CheckState.Checked))

        self.temporal_plot_dock = IconDockWidget(
            "Temporal profile", self, QIcon('icons:temporal.svg'))
        self.temporal_plot_dock.setAllowedAreas(Qt.DockWidgetArea.RightDockWidgetArea)
        self.temporal_plot_dock.setWidget(self.temporal_plot_window)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.temporal_plot_dock)
        # spatial window
        self.spatial_plot_window = SpatialPlotWindow(self.plot_model)
        self.spatial_plot_dock = IconDockWidget(
            "Spatial profile", self, QIcon('icons:spatial.svg'))
        self.spatial_plot_dock.setAllowedAreas(Qt.DockWidgetArea.RightDockWidgetArea)
        self.spatial_plot_dock.setWidget(self.spatial_plot_window)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.spatial_plot_dock)
        # layout
        self.tabifyDockWidget(self.temporal_plot_dock, self.spatial_plot_dock)
        self.temporal_plot_dock.hide()
        self.spatial_plot_dock.hide()

        # Main layout:
        self.main_widget = QSplitter(Qt.Orientation.Horizontal)
        self.main_widget.setChildrenCollapsible(False)
        self.map_band_slider_layout = QVBoxLayout()
        self.map_band_slider_layout.addWidget(self.band_slider)
        self.map_band_slider_layout.addWidget(self.map_widget, stretch=1)
        self.map_band_slider_widget = QWidget()
        self.map_band_slider_widget.setLayout(self.map_band_slider_layout)
        self.main_widget.addWidget(self.map_band_slider_widget)
        self.main_widget.addWidget(self.colormap_widget)
        self.main_widget.setCollapsible(self.main_widget.indexOf(self.colormap_widget), True)
        self.setCentralWidget(self.main_widget)
        self.main_widget.resize(250, 250)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self.layer_dockwidget)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self.minimap_dock_widget)

        # Status bar / footer
        # Logo & version
        self.logo_widget = QLabel(self)
        pix = QPixmap('icons:logo_insarviz.png')
        self.logo_widget.setPixmap(
            pix.scaled(24, 24, Qt.AspectRatioMode.KeepAspectRatio,
                       Qt.TransformationMode.SmoothTransformation))
        logo_text_widget = QLabel(f"InsarViz v.{version.__version__}", self)
        # logo_text_widget.setFont(QFont("Arial", 15, QFont.Bold))
        logo_layout = QHBoxLayout()
        logo_layout.addWidget(self.logo_widget)
        logo_layout.addWidget(logo_text_widget)
        logoandtext_widget = QWidget()
        logoandtext_widget.setLayout(logo_layout)
        self.statusBar().addPermanentWidget(logoandtext_widget)
        # Point info
        self.info_widget = QLabel('x=  , y=  ,z=  ')
        self.info_widget.setMargin(3)
        self.plot_model.updated_pointer_info.connect(self.update_cursor_info)
        self.statusBar().addWidget(self.info_widget)

        log_messages.new_record.connect(self._on_new_log_message)

        # selection toolbar
        self.plotting_toolbar = QToolBar("Selection toolbar")
        self.plotting_toolbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)

        self.inter_action = QAction("Interactive", self)
        self.inter_action.setToolTip("Interactive (Alt+1)")
        self.inter_action.setIcon(QIcon('icons:cursor.png'))
        self.inter_action.triggered.connect(lambda x:
                                            self.map_widget.set_interaction(MapView.INTERACTIVE)
                                            if x else None)
        self.inter_action.setCheckable(True)
        self.inter_action.trigger()
        self.inter_action.setShortcut('Alt+1')

        self.points_action = QAction("Point", self)
        self.points_action.setToolTip("Point (Alt+2)")
        self.points_action.setIcon(QIcon('icons:points.png'))
        self.points_action.setIconText("Point")
        self.points_action.setCheckable(True)
        self.points_action.toggled.connect(lambda x: self.map_widget.set_interaction(MapView.POINTS)
                                           if x else None)
        self.points_action.setShortcut('Alt+2')

        self.prof_action = QAction("Profile", self)
        self.prof_action.setToolTip("Profile (Alt+3)")
        self.prof_action.setIcon(QIcon('icons:profile.png'))
        self.prof_action.setIconText("Profile")
        self.prof_action.setCheckable(True)
        self.prof_action.toggled.connect(lambda x: self.map_widget.set_interaction(MapView.PROFILE)
                                         if x else None)
        self.prof_action.setShortcut('Alt+3')

        self.ref_action = QAction("Reference", self)
        self.ref_action.setToolTip("Reference (Alt+4)")
        self.ref_action.setIcon(QIcon('icons:ref.png'))
        self.ref_action.setIconText("Reference")
        self.ref_action.setCheckable(True)
        self.ref_action.toggled.connect(lambda x: self.map_widget.set_interaction(MapView.REF)
                                        if x else None)
        self.ref_action.setShortcut('Alt+4')

        # action group so that only one tool can be selected at the same time
        self.action_group = QActionGroup(self)
        self.action_group.setExclusive(True)
        self.action_group.addAction(self.inter_action)
        self.action_group.addAction(self.points_action)
        self.action_group.addAction(self.prof_action)
        self.action_group.addAction(self.ref_action)
        self.action_group.setDisabled(True)
        # spacers to center buttons:
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        spacer2 = QWidget()
        spacer2.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.plotting_toolbar.addWidget(spacer)
        self.plotting_toolbar.addAction(self.inter_action)
        self.plotting_toolbar.addAction(self.points_action)
        self.plotting_toolbar.addAction(self.prof_action)
        self.plotting_toolbar.addAction(self.ref_action)
        self.plotting_toolbar.addWidget(spacer2)
        self.addToolBar(self.plotting_toolbar)

        # Menu
        menubar = self.menuBar()
        menubar.setNativeMenuBar(False)

        # File Menu
        file_menu = menubar.addMenu('File')
        open_action = QAction("Open data cube or project", self)
        open_action.setShortcut(QKeySequence.StandardKey.Open)
        open_action.triggered.connect(self.open)

        self.save_action = QAction("Save", self)
        self.save_action.triggered.connect(self.save)
        self.save_action.setShortcut(QKeySequence.StandardKey.Save)
        self.save_action.setDisabled(True)
        self.save_as_project_action = QAction("Save as project", self)
        self.save_as_project_action.triggered.connect(self.save_as_project)
        self.save_as_project_action.setShortcut(QKeySequence.StandardKey.SaveAs)
        self.save_as_project_action.setDisabled(True)

        quit_action = QAction("Quit", self)
        quit_action.triggered.connect(self.close)
        quit_action.setShortcut(QKeySequence.StandardKey.Quit)

        file_menu.addAction(open_action)
        file_menu.addSeparator()
        file_menu.addAction(self.save_action)
        file_menu.addAction(self.save_as_project_action)
        file_menu.addSeparator()
        file_menu.addAction(quit_action)

        # Edit Menu
        edit_menu = menubar.addMenu('Edit')
        undo_action = self.undo_stack.createUndoAction(self, "Undo")
        undo_action.setIcon(QIcon("icons:undo.svg"))
        undo_action.setShortcuts(QKeySequence.StandardKey.Undo)

        redo_action = self.undo_stack.createRedoAction(self, "Redo")
        redo_action.setIcon(QIcon("icons:redo.svg"))
        redo_action.setShortcuts(QKeySequence.StandardKey.Redo)

        edit_menu.addAction(undo_action)
        edit_menu.addAction(redo_action)

        # Layer Menu
        layer_menu = menubar.addMenu('Layer')
        layer_menu.addAction(self.layer_moveup_action)
        layer_menu.addAction(self.layer_movedown_action)
        layer_menu.addAction(self.layer_showall_action)
        layer_menu.addAction(self.layer_hideall_action)
        layer_menu.addAction(self.layer_remove_action)
        layer_menu.addSeparator()
        layer_menu.addAction(self.new_swipe_action)
        layer_menu.addSeparator()
        layer_menu.addAction(self.new_openstreetmap_action)
        layer_menu.addAction(self.new_WMTS_action)
        layer_menu.addAction(self.new_XYZ_action)
        layer_menu.addAction(self.new_raster1B_action)
        layer_menu.addAction(self.new_rasterRGB_action)

        # View Menu
        view_menu = menubar.addMenu('View')
        self.plot_act = QAction("Plotting", self)
        self.plot_act.setCheckable(True)
        self.plot_act.setChecked(False)
        self.plot_act.setEnabled(False)
        self.plot_act.toggled.connect(self.show_plot_window)
        self.plot_act.setShortcut('Ctrl+P')
        view_menu.addAction(self.plot_act)

        self.minimap_action = QAction("Minimap", self)
        self.minimap_action.setCheckable(True)
        self.minimap_action.setChecked(True)
        self.minimap_action.toggled.connect(self.minimap_dock_widget.setVisible)
        self.minimap_dock_widget.visibilityChanged.connect(self.minimap_action.setChecked)
        view_menu.addAction(self.minimap_action)

        self.flip_h_action = QAction("Flip Horizontally", self)
        self.flip_h_action.setToolTip("Flip map horizontally")
        self.flip_h_action.setCheckable(True)
        self.flip_h_action.setChecked(False)
        self.flip_h_action.toggled.connect(self.map_model.set_flip_h)
        self.flip_v_action = QAction("Flip Vertically", self)
        self.flip_v_action.setToolTip("Flip map vertically")
        self.flip_v_action.setCheckable(True)
        self.flip_v_action.setChecked(False)
        self.flip_v_action.toggled.connect(self.map_model.set_flip_v)
        self.flip_group = QActionGroup(self)
        self.flip_group.addAction(self.flip_h_action)
        self.flip_group.addAction(self.flip_v_action)
        self.flip_group.setExclusive(False)
        self.flip_group.setDisabled(True)
        view_menu.addSeparator()
        view_menu.addAction(self.flip_h_action)
        view_menu.addAction(self.flip_v_action)

        # Help Menu
        help_menu = menubar.addMenu('Help')
        help_action = QAction("Documentation", self)
        help_action.triggered.connect(openUrl)
        help_action.setShortcut(QKeySequence.StandardKey.HelpContents)
        help_menu.addAction(help_action)
        show_log_action = QAction("Show InsarViz logs", self)
        show_log_action.triggered.connect(self.show_log_messages_dialog)
        help_menu.addAction(show_log_action)

    @Slot()
    def show_log_messages_dialog(self):
        dialog = QDialog(self)
        dialog.resize(600, 400)
        logger_widget = LoggerWidget(log_messages)
        layout = QVBoxLayout(dialog)
        layout.addWidget(logger_widget)
        dialog.setModal(True)
        dialog.setWindowTitle("InsarViz logs")
        dialog.setLayout(layout)
        dialog.show()

    @Slot()
    def _on_new_log_message(self):
        last_record = log_messages.records[-1]
        if last_record.levelno >= logging.ERROR:
            QMessageBox.critical(self, "Insarviz - Critical error", last_record.msg)

    def load_data(self, filepath_string: str) -> bool:
        # if unsaved changes propose to save
        if not self.undo_stack.isClean():
            changes_managed = False
            while not changes_managed:
                user_input = QMessageBox.warning(self, "Changes not saved",
                                                 "Unsaved changes.\nDo you want to save them inside a project ?",
                                                 QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel,
                                                 QMessageBox.StandardButton.Save)
                if user_input == QMessageBox.StandardButton.Save:
                    if self.save_as_project():
                        changes_managed = True
                elif user_input == QMessageBox.StandardButton.Discard:
                    changes_managed = True
                elif user_input == QMessageBox.StandardButton.Cancel:
                    return False
        # close current project
        self.map_model.close()
        # open new project or cube
        filepath = normalize_path(filepath_string)
        logger.info(f"loading {filepath}")
        extension = filepath.suffix
        if extension == ".json":
            #  opening insarviz project
            input_dict: dict[str, Any]
            with open(filepath, "r", encoding="utf-8") as file:
                try:
                    input_dict = json.load(file)
                except json.JSONDecodeError:
                    logger.warning(f"JSON decoder error on {filepath}")
                    QMessageBox.critical(self, "File Open Error",
                                         f"{filepath}\nJSON decoder error (not a correct JSON file).")
                    return False
            try:
                logger.info(f"{filepath} is an insarviz project version {input_dict['insarviz']}")
            except KeyError:
                logger.warning(f"{filepath} is not an insarviz project")
                QMessageBox.critical(self, "File Open Error",
                                     f"{filepath}\nis not an insarviz project.")
                return False
            try:
                self.loader.open(normalize_path(input_dict["dataset_path"], filepath.as_posix()))
            except rasterio.errors.RasterioIOError as e:
                logger.warning(repr(e))
                QMessageBox.critical(self, "File Open Error", repr(e))
                return False
            try:
                if not self.map_model.from_dict(input_dict, filepath, self.map_widget):
                    logger.warning(
                        f"{filepath} is an insarviz project but its structure is not correct")
                    self.undo_stack.clear()
                    QMessageBox.critical(self, "File Open Error",
                                         f"{filepath}\nis an insarviz project but its structure is not correct")
                    return False
            except rasterio.errors.RasterioIOError as e:
                logger.warning(repr(e))
                self.map_model.close()
                QMessageBox.critical(self, "File Open Error", repr(e))
                return False
            self.project_filepath = filepath
        else:
            # opening cube directly (without project)
            try:
                self.loader.open(filepath)
            except rasterio.errors.RasterioIOError:
                logger.warning(f"gdal cannot open {filepath}")
                QMessageBox.critical(self, "File Open Error",
                                     f"Rasterio / GDAL cannot open:\n{filepath}")
                return False
            self.map_model.create_base_layers()
        logger.info(f"{filepath} successfully loaded")
        # enable plot button in menu:
        self.plot_act.setEnabled(True)
        self.undo_stack.clear()
        self.map_model.opened.emit()
        if self.loader.dataset.crs is None:
            self.geolocated_action_group.setDisabled(True)
        else:
            self.geolocated_action_group.setEnabled(True)
        self.band_slider.set_value(self.map_model.current_band_index)
        self.band_slider.set_reference_band_index(self.loader.reference_band_index)
        self.band_slider.setEnabled(True)
        self.setWindowTitle(f"{str(filepath)}[*] - Insarviz")
        return True

    @Slot(int)
    def on_mapview_interaction_changed(self, value: int) -> None:
        if value == MapView.INTERACTIVE:
            self.inter_action.setChecked(True)
        elif value == MapView.POINTS:
            self.points_action.setChecked(True)
        elif value == MapView.PROFILE:
            self.prof_action.setChecked(True)
        elif value == MapView.REF:
            self.ref_action.setChecked(True)

    @Slot(tuple, np.ndarray)
    def update_cursor_info(self, info: tuple, _: np.ndarray) -> None:
        """
        update point information (x, y, value) in Informations widget as
        cursor hovers over Map

        Parameters
        ----------
        info : tuple (x, y, date, value)
            values of the point currently hovered over in MapView.
        """
        if info == ():
            self.info_widget.setText("")
        else:
            assert self.loader.dataset is not None
            x, y, date, value = info
            if isinstance(date, int):
                date = f"band #{date}"
            elif isinstance(date, datetime.datetime):
                date = date.date()
            text = f"x:{x:5d}, y:{y:5d}, val:{value:7.3f}, date:{date}"
            if self.loader.dataset.crs is not None:
                x_coord, y_coord = self.loader.dataset.xy(y, x)
                long, lat = rasterio.warp.transform(self.loader.dataset.crs,
                                                    rasterio.crs.CRS.from_epsg(4326),
                                                    [x_coord], [y_coord])
                text = text + f", lat:{lat[0]:7.3f}, long:{long[0]:7.3f}"
            self.info_widget.setText(text)

    def show_plot_window(self, checked: bool) -> None:
        if checked:
            self.temporal_plot_dock.show()
            self.spatial_plot_dock.show()
        else:
            self.temporal_plot_dock.hide()
            self.spatial_plot_dock.hide()

    def closeEvent(self, event) -> None:
        if not self.undo_stack.isClean():
            changes_managed = False
            while not changes_managed:
                user_input = QMessageBox.warning(self, "Changes not saved",
                                                 "Unsaved changes.\nDo you want to save them inside a project ?",
                                                 QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel,
                                                 QMessageBox.StandardButton.Save)
                if user_input == QMessageBox.StandardButton.Save:
                    if self.save_as_project():
                        changes_managed = True
                elif user_input == QMessageBox.StandardButton.Discard:
                    changes_managed = True
                elif user_input == QMessageBox.StandardButton.Cancel:
                    event.ignore()
                    return
        self.map_model.close()
        print('\n *** Thank you for using InsarViz, see you soon! *** \n')
        super().closeEvent(event)

    def new_rasterRGB(self) -> None:
        # the second return value of QFileDialog.getOpenFileName is the filter selected by user
        filepath_string, _ = QFileDialog.getOpenFileName(self, "Select file")
        filepath = normalize_path(filepath_string)
        if filepath:
            with warnings.catch_warnings():
                # ignore RuntimeWarning for slices that contain only nans
                warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                        message="Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.")
                try:
                    with rasterio.open(filepath) as file:
                        assert ((file.crs is not None and self.loader.dataset.crs is not None) or
                                (file.shape == self.loader.dataset.shape))
                        input_bands = [(f"{file.indexes[i]}: {file.descriptions[i]}"
                                        if file.descriptions[i] is not None else
                                        f"{file.indexes[i]}") for i in range(file.count)]
                        dialog = RasterLayerDialog(input_bands, RasterRGBLayer.nb_band,
                                                   ["R", "G", "B"], parent=self)
                        dialog.exec()
                        if dialog.result() == QDialog.DialogCode.Accepted:
                            R = file.indexes[dialog.band_comboboxes[0].currentIndex()]
                            G = file.indexes[dialog.band_comboboxes[1].currentIndex()]
                            B = file.indexes[dialog.band_comboboxes[2].currentIndex()]
                            mask = (file.indexes[dialog.mask_combobox.currentIndex()]
                                    if dialog.mask_checkbox.isChecked() else None)
                            self.map_model.layer_model.add_layer(
                                RasterRGBLayer(filepath.name, self.map_model, filepath,
                                               R, G, B, mask))
                except rasterio.errors.RasterioIOError:
                    QMessageBox.critical(self, "File Open Error",
                                         f"Cannot open\n{filepath}\nwith rasterio.")
                except AssertionError:
                    QMessageBox.critical(
                        self, "File Open Error", f"{filepath}\nDoes not have the same size as the dataset and is not geolocalized.")

    def new_raster1B(self) -> None:
        # the second return value of QFileDialog.getOpenFileName is the filter selected by user
        filepath_string, _ = QFileDialog.getOpenFileName(self, "Select file")
        filepath = normalize_path(filepath_string)
        if filepath:
            with warnings.catch_warnings():
                # ignore RuntimeWarning for slices that contain only nans
                warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                        message="Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.")
                try:
                    with rasterio.open(filepath) as file:
                        assert ((file.crs is not None and self.loader.dataset.crs is not None) or
                                (file.shape == self.loader.dataset.shape))
                        if file.count == 1:
                            self.map_model.layer_model.add_layer(
                                Raster1BLayer(filepath.name, self.map_model, filepath, 1, mask=None))
                        else:
                            input_bands = [(f"{file.indexes[i]}: {file.descriptions[i]}"
                                            if file.descriptions[i] is not None else
                                            f"{file.indexes[i]}") for i in range(file.count)]
                            dialog = RasterLayerDialog(input_bands, Raster1BLayer.nb_band,
                                                       ["Band"], parent=self)
                            dialog.exec()
                            if dialog.result() == QDialog.DialogCode.Accepted:
                                b = file.indexes[dialog.band_comboboxes[0].currentIndex()]
                                mask = (file.indexes[dialog.mask_combobox.currentIndex()]
                                        if dialog.mask_checkbox.isChecked() else None)
                                self.map_model.layer_model.add_layer(
                                    Raster1BLayer(filepath.name, self.map_model, filepath, b, mask))
                except rasterio.errors.RasterioIOError:
                    QMessageBox.critical(self, "File Open Error",
                                         f"Cannot open\n{filepath}\nwith rasterio.")
                except AssertionError:
                    QMessageBox.critical(
                        self, "File Open Error", f"{filepath}\nDoes not have the same size as the dataset and is not geolocalized.")

    def new_openstreetmap_layer(self) -> None:
        new_layer = OpenStreetMapLayer(self.map_model, self.map_widget)
        self.map_model.layer_model.add_layer(new_layer)

    def new_WMTS_layer(self) -> None:
        dialog = WMTSLayerDialog(parent=self)
        dialog.exec()
        if dialog.result() == QDialog.DialogCode.Accepted:
            server = dialog.server_input.text()
            layer = dialog.layer_input.text()
            tilematrixset = dialog.tilematrixset_input.text()
            try:
                new_layer = WMTSLayer(layer, self.map_model, self.map_widget,
                                      server, layer, tilematrixset)
            except (RuntimeError, KeyError) as e:
                logger.warning(repr(e))
                return
            self.map_model.layer_model.add_layer(new_layer)

    def new_XYZ_layer(self) -> None:
        dialog = XYZLayerDialog(parent=self)
        dialog.exec()
        if dialog.result() == QDialog.DialogCode.Accepted:
            provider = dialog.provider
            try:
                new_layer = XYZLayer(provider.name, self.map_model, self.map_widget, provider)
            except (RuntimeError, KeyError) as e:
                logger.warning(repr(e))
                return
            self.map_model.layer_model.add_layer(new_layer)

    def new_swipe_layer(self) -> None:
        self.map_model.layer_model.add_layer(SwipeTool(dataset = self.loader.dataset))

    def open(self) -> None:
        # the second return value of QFileDialog.getOpenFileName is the filter selected by user
        filepath, _ = QFileDialog.getOpenFileName(self, "Select file")
        if filepath:
            self.load_data(filepath)

    def save_as_project(self) -> bool:
        filter_string = "Insarviz project (*.json)"
        # the second return value of QFileDialog.getOpenFileName is the filter selected by user
        filepath_string, _ = QFileDialog.getSaveFileName(self, "Save insarviz project to ...",
                                                         filter=filter_string)
        if filepath_string:
            filepath = pathlib.Path(filepath_string)
            extension = filepath.suffix
            if extension == "":
                filepath = filepath.with_suffix(".json")
            self.project_filepath = filepath
            return self.save()
        return False

    def save(self) -> bool:
        if self.project_filepath is not None:
            try:
                extension = self.project_filepath.suffix
                if extension != ".json":
                    raise ValueError
                with open(self.project_filepath, "w", encoding="utf-8") as file:
                    json.dump(self.map_model.to_dict(self.project_filepath), file, indent=4)
                self.undo_stack.setClean()
                logger.info(f"Project saved at {self.project_filepath}")
                self.setWindowTitle(f"{str(self.project_filepath)}[*] - Insarviz")
                return True
            except ValueError:
                error_message = f"Cannot save project at:\n{self.project_filepath}\nWrong file extension (should be .json)."
                logger.warning(error_message)
                QMessageBox.critical(self, "Save Error", error_message)
                return False
            except FileNotFoundError:
                error_message = f"Cannot save project at:\n{self.project_filepath}\nFileNotFoundError."
                logger.critical(error_message)
                QMessageBox.critical(self, "Save Error", error_message)
                return False
        return False

    def on_clean_changed(self, v: bool) -> None:
        self.setWindowModified(not v)
        if not v and self.project_filepath:
            self.save_action.setEnabled(True)
        else:
            self.save_action.setDisabled(True)

    def on_map_model_opened(self) -> None:
        self.layer_action_group.setEnabled(True)
        self.action_group.setEnabled(True)
        self.flip_group.setEnabled(True)
        self.save_as_project_action.setEnabled(True)

    def on_map_model_closed(self) -> None:
        self.project_filepath = None
        self.layer_action_group.setDisabled(True)
        self.geolocated_action_group.setDisabled(True)
        self.action_group.setDisabled(True)
        self.flip_group.setDisabled(True)
        self.save_action.setDisabled(True)
        self.save_as_project_action.setDisabled(True)


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser(description="insar timeseries visualisation")
    parser.add_argument("-v", type=int, default=3,
                        help=("set logging level:"
                              "0 critical, 1 error, 2 warning,"
                              "3 info, 4 debug, default=info"))
    parser.add_argument("-i", type=str, default=None, help="input filepath")
    parser.add_argument("-p", type=str, default=None,
                        help="directory that contains user defined plugins")
    args = parser.parse_args()
    logging_translate = [logging.CRITICAL,
                         logging.ERROR,
                         logging.WARNING,
                         logging.INFO,
                         logging.DEBUG]
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                        level=logging_translate[args.v])
    config = None
    if args.p:
        config = {"plugin_directory": args.p}
        logger.info(f"adding {args.p} as plugin_directory")
    # OpenGL widgets share context even belonging in different windows
    QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts)
    # set OpenGL profile
    opengl_format: QSurfaceFormat = QSurfaceFormat.defaultFormat()
    opengl_format.setVersion(4, 1)
    opengl_format.setProfile(QSurfaceFormat.OpenGLContextProfile.CoreProfile)
    opengl_format.setSwapBehavior(QSurfaceFormat.SwapBehavior.DoubleBuffer)
    opengl_format.setOption(QSurfaceFormat.FormatOption.DeprecatedFunctions, False)
    QSurfaceFormat.setDefaultFormat(opengl_format)
    # set pyqtgraph options
    pg.setConfigOption("background", 'w')
    pg.setConfigOption("foreground", 'k')
    # replace csv exporter by custom csv exporter:
    pg.exporters.Exporter.Exporters.pop(
        pg.exporters.Exporter.Exporters.index(
            pyqtgraph.exporters.CSVExporter))
    pg.exporters.Exporter.Exporters.insert(0, myCSVExporter)
    # add icons directory to Qt paths
    script_dir: pathlib.Path = pathlib.Path(__file__).parent  # directory of ts_viz.py
    icons_dir: pathlib.Path = script_dir / "icons"
    QDir.addSearchPath('icons', str(icons_dir))
    # create application
    app: QApplication = QApplication([])
    main_window = MainWindow(config_dict=config)
    main_window.show()
    # focus on mainwindow
    main_window.showMaximized()
    main_window.activateWindow()
    # main_window2 = MainWindow(filepath=args.i, config_dict=config)
    # main_window2.show()

    # QTASYNCIO
    # QtAsyncio.run(handle_sigint=True)
    # sys.exit(app.exec())
    # END QTASYNCIO

    # QASYNC
    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)
    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    async def async_main():
        # loading directly if file specified upon app launch:
        if args.i is not None:
            main_window.load_data(args.i)
        await app_close_event.wait()

    with event_loop:
        event_loop.run_until_complete(async_main())
    # END QASYNC
    app.quit()
    queue_listener.stop()
    sys.exit(0)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Loader

This module manages data loading from a file.
Contains the class:
    * Loader
and its methods:
    * open - open data file and store dataset
    * __len__ - length of dataset (# of bands or dates)
    * _dates - list of dates or list of band #
    * load_band - load one band data (all points) from file
    * load_profile - load one point data (all dates) from file
    * get_metadata - get metadata from file if exists
"""

from typing import Optional, Union

import logging
import traceback

import time

import pathlib

import multiprocessing

import datetime

import warnings

import asyncio

from qasync import asyncSlot

import concurrent.futures

import numpy as np

import rasterio
import rasterio.warp

from PySide6.QtCore import QObject, Signal, Slot

from insarviz.utils import normalize_path

logger = logging.getLogger(__name__)


# data ######################################################################


class Loader(QObject):

    # used to find the outliers: values outside
    # [q25 - outlier_threshold*IQR, q75 + outlier_threshold*IQR]
    outlier_threshold = 4  # classicaly 1.5, but a higher value seems nicer

    # emited when a dataset is opened, pass the starting band index
    data_loaded = Signal(int)
    data_units_loaded = Signal(str)
    histograms_computed = Signal()
    computing_histograms = Signal()
    data_invalidated = Signal()
    reference_band_index_changed = Signal(object)

    def __init__(self) -> None:
        super().__init__()
        self.filepath: Optional[pathlib.Path] = None
        self.dataset: Optional[rasterio.DatasetReader] = None
        self.reference_band_index: Optional[int] = None
        self.reference_band: Optional[np.ndarray] = None
        self.metadata: dict[str, str] = {}
        self.dates: Union[list[datetime.datetime], list[int]] = []
        self.timestamps: Optional[np.ndarray] = None  # array of floats representing self.dates
        self.units: str = ""
        self.total_histogram: Optional[tuple[np.ndarray, np.ndarray]] = None
        self.band_histograms: list[tuple[np.ndarray, np.ndarray]] = []
        self.process_manager = multiprocessing.Manager()
        self.cancel_compute_histograms_event: Optional[multiprocessing.Event] = None
        self.compute_histograms_task: Optional[asyncio.Future] = None
        self.process_pool = concurrent.futures.ProcessPoolExecutor(max_workers=1)

    def open(self, filepath: pathlib.Path) -> None:
        """
        Open data file and store dataset.

        Parameters
        ----------
        filepath : str, path
            Name of the file to load (with path).
        """
        self.filepath = normalize_path(filepath)
        with warnings.catch_warnings():
            # ignore RuntimeWarning for slices that contain only nans
            warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                    message="Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.")
            self.dataset = rasterio.open(self.filepath)
        logger.info(f"opened {self.filepath}")
        profile = self.dataset.profile
        logger.info("Dataset profile:\n" +
                    "\n".join([f"\t{attr} {profile[attr]}" for attr in profile]))
        self.get_metadata(self.filepath)
        # value units
        try:
            self.units = self.dataset.tags()['VALUE_UNIT']
            logger.info(f'Value unit "{self.units}" found in dataset VALUE_UNIT tag')
        except KeyError:
            logger.info(f"{self.filepath} missing VALUE_UNIT dataset tag")
            try:
                #  TODO ugly fix because flatsim has problems in its metadata
                self.units = self.metadata['Value_unit'].split(",")[0]
                logger.info(f'value unit "{self.units}" found in .meta file')
            except KeyError:
                self.units = 'Undefined units'
                logger.warn(f'no value unit found, taking "{self.units}" instead')
        self.data_units_loaded.emit(self.units)
        # dates and timestamps
        try:
            self.dates = [datetime.datetime.strptime(str(self.dataset.tags(i)["DATE"]), "%Y%m%d")
                          for i in self.dataset.indexes]
            self.timestamps = np.array([d.timestamp() for d in self.dates], dtype=float)
            logger.info("dates found in DATE band tags")
        except (KeyError, ValueError) as error:
            if isinstance(error, KeyError):
                logger.warn(f"{self.filepath} missing DATE band tags")
            if isinstance(error, ValueError):
                logger.info(f"{self.filepath} DATE band tags are not well formated (%Y%m%d expected"
                            ' as for example "20240827", see '
                            'https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes)')
            try:
                self.dates = [datetime.datetime.strptime(x[-8:], "%Y%m%d")
                              for x in self.dataset.descriptions]
                self.timestamps = np.array([d.timestamp() for d in self.dates], dtype=float)
                logger.info("dates found in band descriptions (last 8 characters)")
            except (ValueError, TypeError):
                logger.warn("no date found, taking band numbers as indexes instead")
                self.dates = self.dataset.indexes
                self.timestamps = np.array(self.dates, dtype=float)
        self.total_histogram = (np.array([]), np.array([]))
        self.band_histograms = [(np.array([]), np.array([])) for i in range(len(self))]
        logger.info(f"number of bands = {len(self)}")
        self.data_loaded.emit(self.dataset.indexes[len(self)//2])

    def close(self) -> None:
        if self.dataset is not None:
            self.cancel_compute_histograms()
            logger.info(f"closing {self.filepath}")
            self.dataset.close()
        self.dataset = None
        self.filepath = None
        self.reference_band_index = None
        self.reference_band = None
        self.metadata = {}
        self.dates = []
        self.timestamps = None
        self.units = ""
        self.total_histogram = None
        self.band_histograms = []

    def __len__(self) -> int:
        """
        Length of dataset = number of bands/dates.

        Returns
        -------
        int
            number of band/dates.
        """
        assert self.dataset is not None, "Loader.len: no dataset opened"
        return len(self.dataset.indexes)

    @asyncSlot(tuple)
    async def compute_histograms(self, binrange: Optional[tuple[float, float]] = None) -> None:
        """
        _summary_

        Parameters
        ----------
        bands : Optional[Union[int], list[int]], optional
            Indexes of dataset bands to sample, by default None meaning all bands.
        """
        assert self.dataset is not None, "Loader.compute_histogram: no dataset opened"
        assert self.filepath is not None
        self.computing_histograms.emit()
        self.cancel_compute_histograms_event = self.process_manager.Event()
        loop = asyncio.get_running_loop()
        self.compute_histograms_task = loop.run_in_executor(self.process_pool,
                                                            compute_histograms_worker,
                                                            *(self.filepath,
                                                              self.cancel_compute_histograms_event,
                                                              binrange))
        try:
            await self.compute_histograms_task
        except Exception as e:
            logger.error(traceback.format_exc())

        self.cancel_compute_histograms_event = None
        if not self.compute_histograms_task.cancelled():
            self.set_histograms(*self.compute_histograms_task.result())
        else:
            self.set_histograms(self.total_histogram, self.band_histograms)
        self.compute_histograms_task = None

    def cancel_compute_histograms(self):
        if self.total_histogram is None or len(self.total_histogram) == 0:
            return
        if self.compute_histograms_task is not None:
            self.compute_histograms_task.cancel()
            if self.cancel_compute_histograms_event is not None:
                self.cancel_compute_histograms_event.set()

    def set_histograms(self, total_histogram: tuple[np.ndarray, np.ndarray],
                       band_histograms: list[tuple[np.ndarray, np.ndarray]]) -> bool:
        assert self.dataset is not None
        if len(band_histograms) != len(self):
            logger.warning("number of band histograms not equal to number of bands")
            return False
        total_histogram = tuple([np.array(a) for a in total_histogram])
        if len(total_histogram[0].shape) != 1 or len(total_histogram[1].shape) != 1:
            logger.warning("total histogram is not an array of dimension 1")
            return False
        if len(total_histogram[1]) != len(total_histogram[0]) + 1:
            logger.warning("total histogram number of bins is not equal to number of values + 1")
            return False
        for i in range(len(self)):
            band_histograms[i] = tuple([np.array(a) for a in band_histograms[i]])
            if len(band_histograms[i][0].shape) != 1 or len(band_histograms[i][1].shape) != 1:
                logger.warning(f"band {self.dataset.indexes[i]} histogram is not an array of "
                               "dimension 1")
                return False
            if len(band_histograms[i][1]) != len(band_histograms[i][0]) + 1:
                logger.warning(
                    f"band {self.dataset.indexes[i]} histogram number of bins is not equal to "
                    "number of values + 1")
                return False
        self.total_histogram = total_histogram
        self.band_histograms = band_histograms
        self.histograms_computed.emit()
        return True

    def load_band(self, i: int, use_reference: bool = True) -> np.ndarray:
        """
        load band i from dataset

        Parameters
        ----------
        i : int, optional
            Band number to load. The default is 0.

        Returns
        -------
        band : array
            Loaded band data.
        """
        assert self.dataset is not None, "Loader.load_band: no dataset opened"
        assert i in self.dataset.indexes
        t0 = time.time()
        band = self.dataset.read(i)
        try:
            no_data = self.dataset.nodatavals[i]
        except (IndexError, AttributeError):
            try:
                no_data = self.dataset.nodata
            except AttributeError:
                no_data = np.nan
        if no_data is not np.nan:
            band[band == no_data] = np.nan
        if use_reference:
            if self.reference_band_index is not None and self.reference_band is not None:
                band = band - self.reference_band
        t1 = time.time()
        logger.info(f"loaded band {i} in {t1-t0} s")
        return band

    def load_profile_window(self, i_start: float, j_start: float, i_stop: Optional[float] = None,
                            j_stop: Optional[float] = None) -> np.ndarray:
        assert self.dataset is not None, "Loader.load_profile_window: no dataset opened"
        i, j = int(i_start), int(j_start)
        width = abs(i_start - int(i_stop)) + 1 if i_stop else 1
        length = abs(j_start - int(j_stop)) + 1 if j_stop else 1
        data = self.dataset.read(self.dataset.indexes,
                                 window=rasterio.windows.Window(i, j, width, length))
        for index in range(len(self)):
            try:
                no_data = self.dataset.nodatavals[self.dataset.indexes[index]]
            except (IndexError, AttributeError):
                try:
                    no_data = self.dataset.nodata
                except AttributeError:
                    continue
            data[index][data[index] == no_data] = np.nan
        if self.reference_band_index is not None and self.reference_band is not None:
            data = data - self.reference_band[j:j+length, i:i+width]
        with warnings.catch_warnings():
            # ignore RuntimeWarning for slices that contain only nans
            warnings.filterwarnings("ignore", category=RuntimeWarning,
                                    message="Mean of empty slice")
            data = np.nanmean(data, axis=(1, 2))
        return data

    @Slot(object)
    def set_reference_band(self, index: Optional[int]) -> None:
        if index is not None:
            index = int(index)
        if self.reference_band_index != index:
            self.reference_band_index = index
            if self.reference_band_index is None:
                self.reference_band = None
            else:
                self.reference_band = self.load_band(self.reference_band_index, use_reference=False)
            logger.info(f"reference band changed: {self.reference_band_index}")
            self.data_invalidated.emit()
            self.reference_band_index_changed.emit(self.reference_band_index)

    def get_metadata(self, filename: pathlib.Path) -> None:
        """
        Create a dictionnary containing all metadata entries in self.matadata,
        if a '.meta' file exists in same repo as the datacube file

        Parameters
        ----------
        filename : pathlib.Path
            name of the datacube file
        """
        metafilename = filename.with_suffix('.meta')
        self.metadata = {}
        try:
            with open(metafilename, encoding="utf-8") as f:
                for line in f:
                    try:
                        (key, val) = line.split(sep=': ', maxsplit=1)
                        self.metadata[key] = val.strip()
                    except ValueError:
                        pass
        except FileNotFoundError:
            logger.info("no metadata file found")


# COMPUTE HISTOGRAMS WORKER FUNCTION ###############################################################
def compute_histograms_worker(filepath: pathlib.Path, stop_event: multiprocessing.Event,
                              binrange: Optional[tuple[float, float]] = None):
    dataset = rasterio.open(filepath)
    # intialize progress dialog
    t0 = time.time()
    # xy is an array of (row, col, band) positions to randomly sample the dataset
    sample_size = 20000
    xy = np.random.randint([dataset.height, dataset.width, len(dataset.indexes)],
                           size=(sample_size, 3))

    # sort xy to optimize dataset sampling (split between interleaving cases)
    if len(set(dataset.block_shapes)) == 1:
        # block shape is the same across bands
        block_height, block_width = dataset.block_shapes[0]
        if (dataset.interleaving == rasterio.enums.Interleaving.band
                or (dataset.driver == 'VRT' and dataset.interleaving is None)):
            # sort xy first by band, second by block row, third by block column
            xy = xy[np.lexsort(
                (xy[:, 1] // block_width, xy[:, 0] // block_height, xy[:, 2]))]
        if dataset.interleaving == rasterio.enums.Interleaving.pixel:
            # sort xy first by block row, second by block column
            xy = xy[np.lexsort((xy[:, 1] // block_width, xy[:, 0] // block_height))]
        else:
            # Assume interleave=pixel otherwise
            xy = xy[np.lexsort((xy[:, 1] // block_width, xy[:, 0] // block_height))]

    else:
        # block shape is not the same across bands so interleaving must be band
        assert dataset.interleaving == rasterio.enums.Interleaving.band
        # sort xy by band
        xy = xy[np.lexsort((xy[:, 2]))]
    if stop_event.is_set():
        raise asyncio.CancelledError
    # sample the dataset
    sample = np.empty(sample_size)
    for i in range(20):
        # split in 20 steps to update the progress dialog
        start = (i * sample_size) // 20
        stop = ((i+1) * sample_size) // 20
        sample[start:stop] = np.array([
            dataset.read(dataset.indexes[k[2]],
                         window=rasterio.windows.Window(k[1], k[0], 1, 1))[0][0]
            for k in xy[start:stop]])
        if stop_event.is_set():
            raise asyncio.CancelledError
    logger.debug(f"Nodata value : {dataset.nodata}")
    sample = sample[~np.isnan(sample)] # Remove nans, always
    if dataset.nodata is not None and not np.isnan(dataset.nodata):
        sample = sample[sample != dataset.nodata]
    IQR = 0.0
    logger.debug(f"sample size : {len(sample)}")
    percentileBounds = (25,75)
    while IQR == 0.0:
        # compute the histogram bins on the sample
        # get the first and third quartiles (inverted_cdf means exact observation)
        qMin, qMax = np.percentile(sample, percentileBounds, method="inverted_cdf")
        percentileBounds = (percentileBounds[0]*0.5, 50+percentileBounds[1]*0.5)
        # interquartile range
        IQR = qMax - qMin
    # Freedman Diaconis Estimator of the binwidth
    binwidth = 2 * IQR / (np.cbrt(len(sample)))
    logger.debug(f"IQR : {IQR}")
    if binrange is None:
        # range exclude outliers, ie observations outside of
        # [q25 - outlier_threshold*IQR, q75 + outlier_threshold*IQR]
        binrange = (qMin - Loader.outlier_threshold*IQR,
                    qMax + Loader.outlier_threshold*IQR)
    # compute the bin count as in np.histogram_bin_edges
    bincount = int(np.round(np.ceil((binrange[1] - binrange[0]) / binwidth)))
    # compute the bins
    bins = np.histogram_bin_edges(sample, bins=bincount, range=binrange)
    # add a bin at the start and end for outliers
    bins_total = np.array([bins[0], *bins, bins[-1]])
    bins_band = [np.array([bins[0], *bins, bins[-1]]) for _ in range(len(dataset.indexes))]
    if stop_event.is_set():
        raise asyncio.CancelledError
    # build the histograms (split between interleaving cases)
    nd = dataset.profile['nodata']
    histogram_total = np.zeros(bins_total.shape[0]-1)
    histograms_band = [np.zeros(bins_band[i].shape[0]-1)
                       for i in range(len(dataset.indexes))]
    with warnings.catch_warnings():
        # ignore RuntimeWarning for slices that contain only nans
        warnings.filterwarnings("ignore", category=RuntimeWarning,
                                message="All-NaN slice encountered")
        if dataset.interleaving == rasterio.enums.Interleaving.pixel:
            # pixel interleaving => open block by block
            assert len(set(dataset.block_shapes)) == 1
            block = np.empty((len(dataset.indexes), *dataset.block_shapes[0]))
            block_count = sum(1 for _ in dataset.block_windows())
            for k, window in enumerate(dataset.block_windows()):
                block = dataset.read(window=window[1], out=block)
                block[block == nd] = np.nan
                for i in range(len(dataset.indexes)):
                    tmp_min = np.nanmin(block[i])
                    bins_total[0] = np.nanmin((tmp_min, bins_total[0]))
                    bins_band[i][0] = np.nanmin((tmp_min, bins_band[i][0]))
                    tmp_max = np.nanmax(block[i])
                    bins_total[-1] = np.nanmax((tmp_max, bins_total[-1]))
                    bins_band[i][-1] = np.nanmax((tmp_max, bins_band[i][-1]))
                    tmp_hist = np.histogram(block[i], bins=bins_band[i])[0]
                    histograms_band[i] += tmp_hist
                    histogram_total += tmp_hist
                if stop_event.is_set():
                    raise asyncio.CancelledError
        else:
            # band interleaving or no interleaving metadata => open band by band
            band = np.empty(dataset.shape)
            band_count = len(dataset.indexes)
            for i, idx in enumerate(dataset.indexes):
                band = dataset.read(idx, out=band)
                band[band == nd] = np.nan
                band_min = np.nanmin(band)
                bins_total[0] = np.nanmin((band_min, bins_total[0]))
                bins_band[i][0] = np.nanmin((band_min, bins_band[i][0]))
                band_max = np.nanmax(band)
                bins_total[-1] = np.nanmax((band_max, bins_total[-1]))
                bins_band[i][-1] = np.nanmax((band_max, bins_band[i][-1]))
                histograms_band[i] += np.histogram(band, bins=bins_band[i])[0]
                histogram_total += histograms_band[i]
                if stop_event.is_set():
                    raise asyncio.CancelledError
    # normalize the histograms in order to plot them in ColormapWidget
    for i, h in enumerate(histograms_band):
        histograms_band[i] = h / sum(h)
    histogram_total = histogram_total / sum(histogram_total)
    t1 = time.time()
    logger.info(f"computed histograms in {t1-t0}s ({len(bins)-1} bins)")
    if stop_event.is_set():
        raise asyncio.CancelledError
    return (histogram_total, bins_total), list(zip(histograms_band, bins_band))



# -*- coding: utf-8 -*-

from typing import Any, Optional

from PySide6.QtCore import Qt

from PySide6.QtGui import QColor

from pyqtgraph.colormap import ColorMap

from insarviz.map.TreeModel import TreeItem, TreeModel

from insarviz.Roles import Roles


# TreeItemAttribute class ##########################################################################

class TreeItemAttribute(TreeItem):
    """
    TreeItemAttribute provides user-friendly access to a TreeItem's attribute. Used as a child of
    an item, it represents one of its attribute that can be modified by interacting with the
    TreeItemAttribute. TreeItemAttribute use getattr and setattr to get/set the attribute of its
    parent. LayerModel.setData is responsible to emit a dataChanged signal linked to the
    TreeItemAttribute's parent when the latter is modified.

    TreeItemAttribute.data(..., Roles.EditorRole) returns a string used by LayerView.ItemDelegate
    to create custom editor. TreeItemAttribute.data(..., Roles.DataRole) returns an insarviz.Roles
    included in the dataChanged emited by LayerModel.setData on the TreeItemAttribute's parent,
    this Roles is used for example by PlotModel to know if it has to recompute the data of
    a SelectionItem.

    Abstract class than need to be subclassed.
    """

    EditorRole: int = int(Roles.EditorRole)
    DataRole: int = int(Roles.DataRole)
    PersistentEditorWidgetRole: int = int(Roles.PersistentEditorWidgetRole)

    def __init__(self, parent: TreeItem, attribute_name: str, name: Optional[str] = None,
                 tooltip: Optional[str] = None, datarole: Optional[int] = None,
                 editable: bool = True):
        super().__init__(parent)
        assert parent, "TreeItemAttribute must have valid parent"
        try:
            getattr(parent, attribute_name)
        except AttributeError as e:
            raise AttributeError(f"TreeItemAttribute's parent must have an attribute matching \
                    attribute_name {attribute_name}\n{repr(e)}")
        self.attribute_name: str = attribute_name
        self.name: str
        if name is None:
            self.name = attribute_name
        else:
            self.name = name
        self.tooltip: Optional[str] = tooltip
        self.datarole: Optional[int] = datarole
        self.editable: bool = editable

    def data(self, column: int, role: int) -> Any:
        if column == TreeModel.remove_column:
            return None
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return f"{self.name} : {getattr(self.parent(), self.attribute_name)}"
        if role == Qt.ItemDataRole.ToolTipRole:
            return self.tooltip
        if role == TreeItemAttribute.DataRole:
            # used by LayerView.ItemDelegate to emit custom role in LayerModel.dataChanged
            return int(self.datarole) if self.datarole is not None else None
        if role == TreeItemAttribute.EditorRole:
            # must be implemented by subclasses and return a string that is used by
            # LayerView.ItemDelegate, this string shall not be None for self to be recognized
            # as a TreeItemAttribute in LayerModel
            raise NotImplementedError()
        return None

    def flags(self, flags: Qt.ItemFlags, column: int) -> Qt.ItemFlags:
        if column == 0 and self.editable:
            flags = flags | Qt.ItemFlag.ItemIsEditable
        elif column == TreeModel.remove_column:
            flags = flags & (~Qt.ItemFlag.ItemIsEnabled)
        return flags

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if column != 0:
            return False
        if role == Qt.ItemDataRole.EditRole:
            setattr(self.parent(), self.attribute_name, value)
            return True
        return False


class TreeItemIntegerAttribute(TreeItemAttribute):

    def __init__(self, parent: TreeItem, attribute_name: str, name: Optional[str] = None,
                 tooltip: Optional[str] = None, vmin: Optional[int] = None,
                 vmax: Optional[int] = None, unit: Optional[str] = None,
                 datarole: Optional[int] = None):
        super().__init__(parent, attribute_name, name=name, tooltip=tooltip, datarole=datarole)
        self.vmin: Optional[int] = vmin
        self.vmax: Optional[int] = vmax
        self.unit: Optional[str] = unit

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            if self.unit:
                return f"{self.name} : {getattr(self.parent(), self.attribute_name)} {self.unit}"
        if role == Qt.ItemDataRole.EditRole:
            # (value, name, vmin, vmax, unit)
            return (getattr(self.parent(), self.attribute_name), self.name, self.vmin, self.vmax,
                    self.unit)
        if role == TreeItemAttribute.EditorRole:
            # used by LayerView.ItemDelegate to create custom editor
            return "integer"
        return super().data(column, role)

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if not isinstance(value, int):
            return False
        if self.vmin is not None:
            if value < self.vmin:
                return False
        if self.vmax is not None:
            if value > self.vmax:
                return False
        return super().set_data(value, column, role)


class TreeItemFloatAttribute(TreeItemAttribute):

    def __init__(self, parent: TreeItem, attribute_name: str, name: Optional[str] = None,
                 tooltip: Optional[str] = None, vmin: Optional[float] = None,
                 vmax: Optional[float] = None, unit: Optional[str] = None,
                 datarole: Optional[int] = None):
        super().__init__(parent, attribute_name, name=name, tooltip=tooltip, datarole=datarole)
        self.vmin: Optional[float] = vmin
        self.vmax: Optional[float] = vmax
        self.unit: Optional[str] = unit

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            if self.unit:
                return f"{self.name} : {getattr(self.parent(), self.attribute_name)} {self.unit}"
        if role == Qt.ItemDataRole.EditRole:
            # (value, name, vmin, vmax, unit)
            return (getattr(self.parent(), self.attribute_name), self.name, self.vmin, self.vmax,
                    self.unit)
        if role == TreeItemAttribute.EditorRole:
            # used by LayerView.ItemDelegate to create custom editor
            return "float"
        return super().data(column, role)

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if not isinstance(value, float):
            return False
        if self.vmin is not None:
            if value < self.vmin:
                return False
        if self.vmax is not None:
            if value > self.vmax:
                return False
        return super().set_data(value, column, role)

class TreeItemSliderAttribute(TreeItemAttribute):

    def __init__(self, parent: TreeItem,
                 attribute_name: str, name: Optional[str] = None,
                 tooltip: Optional[str] = None, vmin: Optional[float] = None,
                 vmax: Optional[float] = None, datarole: Optional[int] = None):
        super().__init__(parent, attribute_name, name=name, tooltip=tooltip, datarole=datarole)
        self.vmin: Optional[float] = vmin
        self.vmax: Optional[float] = vmax
        self.persistent_editor = None

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            # Having None as a display role means the ItemDelegate
            # will create a persistent editor instead
            return None
        if role == Qt.ItemDataRole.EditRole:
            # (value, name, vmin, vmax)
            return (getattr(self.parent(), self.attribute_name), self.name, self.vmin, self.vmax)
        if role == Qt.ItemDataRole.SizeHintRole:
            if self.persistent_editor is not None:
                return self.persistent_editor.sizeHint()
        if role == TreeItemAttribute.EditorRole:
            # used by LayerView.ItemDelegate to create custom editor
            return "float_slider"
        if role == TreeItemAttribute.PersistentEditorWidgetRole:
            return self.persistent_editor
        return super().data(column, role)

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if role == TreeItemAttribute.PersistentEditorWidgetRole:
            if self.persistent_editor is None:
                self.persistent_editor = value
                return True
            else:
                return False

        if not isinstance(value, float):
            return False
        if self.vmin is not None:
            if value < self.vmin:
                return False
        if self.vmax is not None:
            if value > self.vmax:
                return False

        return super().set_data(value, column, role)

class TreeItemColorAttribute(TreeItemAttribute):

    def data(self, column: int, role: int) -> Any:
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return getattr(self.parent(), self.attribute_name)
        if role == TreeItemAttribute.EditorRole:
            # used by LayerView.ItemDelegate to create custom editor
            return "color"
        return super().data(column, role)

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if not isinstance(value, QColor):
            return False
        return super().set_data(value, column, role)


class TreeItemColormapAttribute(TreeItemAttribute):

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            return getattr(self.parent(), self.attribute_name)
        if role == Qt.ItemDataRole.EditRole:
            return self.parent()
        if role == TreeItemAttribute.EditorRole:
            # used by LayerView.ItemDelegate to create custom editor
            return "colormap"
        return super().data(column, role)

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if column != 0:
            return False
        if role == Qt.ItemDataRole.EditRole and isinstance(value, tuple):
            if len(value) != 3:
                return False
            colormap, v0, v1 = value
            if ((not isinstance(colormap, ColorMap)) or (not isinstance(v0, float)) or
                    (not isinstance(v1, float))):
                return False
            self.parent().set_colormap(colormap)
            self.parent().set_v0_v1(v0, v1)
            return True
        return False



# -*- coding: utf-8 -*-

from typing import Optional, Union

import warnings

import datetime

import numpy as np

import pyqtgraph as pg

from PySide6.QtCore import Qt, Slot, Signal, QModelIndex, QPointF

from PySide6.QtWidgets import QComboBox, QLabel, QCheckBox, QBoxLayout, QWidget, QSizePolicy

from PySide6.QtGui import QIcon, QColor, QGuiApplication

from insarviz.plot.PlotModel import PlotModel

from insarviz.plot.AbstractPlotView import AbstractPlotWindow, AbstractPlotWidget, ProxyNoItemModel

from insarviz.BandSlider import BandSlider

from insarviz.Roles import Roles

from insarviz.utils import get_nearest


# SpatialPlotWidget class ##########################################################################

class SpatialPlotWidget(AbstractPlotWidget):
    """
    spatial evolution plot (for profile only)
    """

    def __init__(self, plot_model: PlotModel):
        super().__init__(plot_model)
        # used by insarviz.exporters.myCSVExporter
        self.plotItem.insarviz_widget = self
        self.plotItem.setLabel('bottom', "Distance along profile line", units='pixel')
        self.point_marker = PointMarker()
        self.point_marker.sigPositionChanged.connect(self.point_marker_pos_changed)
        self.profile_curve: pg.PlotDataItem = pg.PlotDataItem(symbol='o', symbolSize=5,
                                                              antialias=True)
        self.profile_curve.sigPointsHovered.connect(self.on_points_hover)
        self.profile_curve.scatter.setData(hoverable=True, tip=None, hoverBrush='r', hoverSize=8)
        self.profile_curve.setSymbolPen(None)
        self.lowstd_curve = pg.PlotCurveItem(connect='finite', pen=None)
        self.highstd_curve = pg.PlotCurveItem(connect='finite', pen=None)
        self.profile_std: pg.FillBetweenItem = pg.FillBetweenItem(self.lowstd_curve,
                                                                  self.highstd_curve)
        self.profile_std.setZValue(-1.)
        self.profile: int = -1
        self.date: Optional[int] = None  # band number addex (starting from 0 like in python)
        self.reference: int = -1
        self.show_std: bool = False
        # signals and slots
        plot_model.profiles_updated.connect(self.update_curves)

    @Slot(int, int, list)
    def update_curves(self, first: int, last: int, roles: list[int]) -> None:
        if first <= self.profile <= last:
            profile_index = self.plot_model.profiles.index(self.profile, 0)
            if Roles.ComputeDataRole in roles:
                self.compute_profile_curve()
                self.compute_profile_std()
            if Roles.CurveColorRole in roles:
                color = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DecorationRole)
                self.profile_curve.setPen(color, width=2)
                self.profile_curve.setSymbolBrush(color)
                self.profile_std.setBrush(self.create_fill_color(color))
            if Qt.ItemDataRole.EditRole in roles:
                name = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DisplayRole)
                assert self.date is not None and self.plot_model.loader.timestamps is not None
                # remove tooltip for this curve
                self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)
                date = self.plot_model.loader.dates[self.date]
                date_string = (date.strftime('%Y-%m-%d') if isinstance(date, datetime.datetime)
                               else f"band {date}")
                self.profile_curve.opts["name"] = (f"{name} at {date_string}")

    @Slot(int)
    def change_profile(self, i: int) -> None:
        if self.profile != i:
            self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)
            if self.profile != -1:
                profile_index = self.plot_model.profiles.index(self.profile, 0)
                self.plot_model.profiles.setData(profile_index, None, Roles.ProfileSpatialRole)
        self.profile = i
        if self.profile != -1:
            profile_index = self.plot_model.profiles.index(self.profile, 0)
            color = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DecorationRole)
            self.profile_curve.setPen(color, width=2)
            self.profile_curve.setSymbolBrush(color)
            name = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DisplayRole)
            assert self.date is not None and self.plot_model.loader.timestamps is not None
            date = self.plot_model.loader.dates[self.date]
            date_string = (date.strftime('%Y-%m-%d') if isinstance(date, datetime.datetime)
                           else f"band {date}")
            self.profile_curve.opts["name"] = (f"{name} at {date_string}")
            self.compute_profile_curve()
            self.profile_std.setBrush(self.create_fill_color(color))
            self.compute_profile_std()
            self.compute_limits()
            self.plotItem.setMouseEnabled(x=True, y=True)
            self.plotItem.addItem(self.profile_curve)
            if self.show_std:
                self.plotItem.addItem(self.profile_std)
            self.point_marker.setPos(0)
            self.point_marker_pos_changed()
            self.point_marker.length = len(self.plot_model.profiles_data[self.profile])
            self.plotItem.addItem(self.point_marker)
        else:
            # remove tooltip for this curve
            self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)
            self.profile_curve.opts["name"] = "no profile"
            self.plotItem.setMouseEnabled(x=False, y=False)
            self.plotItem.removeItem(self.profile_curve)
            self.plotItem.removeItem(self.profile_std)
            self.plotItem.removeItem(self.point_marker)
            self.point_marker.setPos(-1)
            self.point_marker.length = None

    @Slot(int)
    def change_date(self, i: int) -> None:
        assert self.plot_model.loader.dataset is not None
        assert i in self.plot_model.loader.dataset.indexes
        self.date = self.plot_model.loader.dataset.indexes.index(i)
        if self.profile != -1:
            self.compute_profile_curve()

    @Slot(int)
    def change_reference(self, i: int) -> None:
        self.reference = i
        if self.profile != -1:
            self.compute_limits()
            self.compute_profile_curve()
            self.compute_profile_std()

    # connected to SpatialPlotWindow.checkbox_std.stateChanged
    @Slot(int)
    def change_show_std(self, show: int) -> None:
        if show == Qt.CheckState.Checked.value:
            self.show_std = True
            if self.profile != -1:
                self.plotItem.addItem(self.profile_std)
        elif show == Qt.CheckState.Unchecked.value:
            self.show_std = False
            self.plotItem.removeItem(self.profile_std)

    def compute_profile_curve(self) -> None:
        """
        Compute the curve for profile.
        """
        if self.reference == -1:
            reference = np.zeros(len(self.plot_model.loader.dates))
        else:
            reference = np.array(self.plot_model.references_data[self.reference])
        length = self.plot_model.profiles_data[self.profile].shape[0]
        x = np.arange(length)
        assert self.date is not None
        y = self.plot_model.profiles_data[self.profile][:, self.date] - reference[self.date]
        self.profile_curve.setData(x, y)
        # remove tooltip for this curve
        self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)

    def compute_profile_std(self) -> None:
        """
        Compute the standard deviation zones (mean +/- std) for profile.
        Set those zones to self.profiles_std.
        """
        if self.reference == -1:
            reference = np.zeros(len(self.plot_model.loader.dates))
        else:
            reference = np.array(self.plot_model.references_data[self.reference])
        x = np.arange(len(self.plot_model.profiles_data[self.profile]))
        with warnings.catch_warnings():
            # ignore RuntimeWarning for slices that contain only nans
            warnings.filterwarnings("ignore", category=RuntimeWarning,
                                    message="Mean of empty slice")
            warnings.filterwarnings("ignore", category=RuntimeWarning,
                                    message="Degrees of freedom <= 0 for slice.")
            profile_tmean = np.nanmean(self.plot_model.profiles_data[self.profile]
                                       - reference, axis=1)
            profile_std = np.nanstd(self.plot_model.profiles_data[self.profile] - reference, axis=1)
        y = profile_tmean - profile_std
        # remove single values surrounded by nans as pg.FillBetweenItem is bugged for those
        y[np.where(np.append(np.isnan(y[1:]), True)
                   & np.append([True], np.isnan(y[:-1])))] = np.nan
        self.lowstd_curve.setData(x, y)
        y = profile_tmean + profile_std
        # remove single values surrounded by nans as pg.FillBetweenItem is bugged for those
        y[np.where(np.append(np.isnan(y[1:]), True)
                   & np.append([True], np.isnan(y[:-1])))] = np.nan
        self.highstd_curve.setData(x, y)

    def compute_limits(self):
        """
        Compute and set axis limits.
        """
        if self.profile != -1:
            length = len(self.plot_model.profiles_data[self.profile])
            x_min = 0. - self.max_padding * length
            x_max = length + self.max_padding * length
            if self.reference == -1:
                reference = np.zeros(len(self.plot_model.loader))
            else:
                reference = np.array(self.plot_model.references_data[self.reference])
            with warnings.catch_warnings():
                # ignore RuntimeWarning for slices that contain only nans
                warnings.filterwarnings("ignore", category=RuntimeWarning,
                                        message="All-NaN slice encountered")
                y_min = np.nanmin(self.plot_model.profiles_data[self.profile] - reference)
                y_max = np.nanmax(self.plot_model.profiles_data[self.profile] - reference)
            y_delta = y_max - y_min
            y_min, y_max = y_min - self.max_padding*y_delta, y_max + self.max_padding*y_delta
            self.plotItem.setLimits(xMin=x_min, xMax=x_max, yMin=y_min, yMax=y_max)
            self.plotItem.setXRange(x_min, x_max, padding=self.default_padding)
            self.plotItem.setYRange(y_min, y_max, padding=self.default_padding)

    def create_fill_color(self, color: QColor) -> QColor:
        fill_color = QColor(color)
        fill_color.setAlpha(96)
        return fill_color

    @Slot(object)
    def point_marker_pos_changed(self, _=None):
        if self.profile != -1:
            profile_index = self.plot_model.profiles.index(self.profile, 0)
            self.plot_model.profiles.setData(profile_index, self.point_marker.getPos()[0],
                                             Roles.ProfileSpatialRole)


# SpatialPlotWindow class ##########################################################################

class SpatialPlotWindow(AbstractPlotWindow):

    reference_changed = Signal(int)
    profile_changed = Signal(int)

    def __init__(self, plot_model: PlotModel):
        super().__init__(plot_model, SpatialPlotWidget(plot_model))
        plot_model.loader.data_loaded.connect(self.on_data_loaded)
        plot_model.closed.connect(self.on_close)
        self.setWindowTitle("Spatial profile")
        # profile icon
        self.toolbar.addSeparator()
        profiles_icon = QLabel()
        profiles_icon.setPixmap(QIcon('icons:profile.png').pixmap(self.toolbar.iconSize()))
        self.toolbar.addWidget(profiles_icon)
        # combobox to choose profile
        self.profile_combobox = QComboBox(self)
        self.toolbar.addWidget(self.profile_combobox)
        proxy_profile_model = ProxyNoItemModel(self.plot_model.profiles, "No profile")
        self.profile_combobox.setModel(proxy_profile_model)
        proxy_profile_model.rowsAboutToBeRemoved.connect(
            self.update_profile_combobox_index_before_remove)
        self.profile_combobox.currentIndexChanged.connect(self.profile_combobox_index_changed)
        self.profile_changed.connect(self.plot_widget.change_profile)
        # reference icon
        self.toolbar.addSeparator()
        reference_icon = QLabel()
        reference_icon.setPixmap(QIcon('icons:ref.png').pixmap(self.toolbar.iconSize()))
        self.toolbar.addWidget(reference_icon)
        # combobox to choose reference
        self.reference_combobox = QComboBox(self)
        self.toolbar.addWidget(self.reference_combobox)
        proxy_reference_model = ProxyNoItemModel(self.plot_model.references, "No reference")
        self.reference_combobox.setModel(proxy_reference_model)
        proxy_reference_model.rowsAboutToBeRemoved.connect(
            self.update_reference_combobox_index_before_remove)
        self.reference_combobox.currentIndexChanged.connect(self.reference_combobox_index_changed)
        self.reference_changed.connect(self.plot_widget.change_reference)
        # checkbox to show standard deviation
        self.toolbar.addSeparator()
        self.checkbox_std: QCheckBox = QCheckBox('Show standard deviation', self)
        self.checkbox_std.setToolTip("When checked, standard deviation across time will be \
                                displayed as a filled zone.")
        self.checkbox_std.stateChanged.connect(self.plot_widget.change_show_std)
        self.toolbar.addWidget(self.checkbox_std)
        # spacer
        self.spacer = QWidget()
        self.spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        self.toolbar.addWidget(self.spacer)
        # export button
        self.toolbar.addWidget(self.export_button)
        # date slider
        self.band_slider = BandSlider(self.plot_model.loader)
        self.band_slider.value_changed.connect(self.plot_widget.change_date)
        # layout
        layout = self.layout()
        assert isinstance(layout, QBoxLayout)
        layout.addWidget(self.band_slider)
        layout.addWidget(self.plot_widget, stretch=1)

    @Slot(int)
    def on_data_loaded(self, band_index: int) -> None:
        self.band_slider.on_data_loaded(band_index)
        assert self.plot_model.loader.dataset is not None
        assert band_index in self.plot_model.loader.dataset.indexes
        self.band_slider.set_value(band_index)

    @Slot()
    def on_close(self) -> None:
        self.profile_combobox.setCurrentIndex(-1)
        self.checkbox_std.setChecked(False)
        self.reference_combobox.setCurrentIndex(-1)
        self.band_slider.on_close()

    @Slot(int)
    def reference_combobox_index_changed(self, i: int) -> None:
        if i == -1:
            # combobox is empty so we reset it with the "no reference" item
            self.reference_combobox.setCurrentIndex(0)
        else:
            # for SpatialPlotWidget.reference, -1 means no reference, 0 means first reference...
            self.reference_changed.emit(i-1)

    @Slot(int)
    def profile_combobox_index_changed(self, i: int) -> None:
        if i == -1:
            # combobox is empty so we reset it with the "no reference" item
            self.profile_combobox.setCurrentIndex(0)
        else:
            if i == 0:
                self.band_slider.setDisabled(True)
            else:
                self.band_slider.setEnabled(True)
            # for SpatialPlotWidget.profile, -1 means no profile, 0 means first profile...
            self.profile_changed.emit(i-1)

    @Slot(QModelIndex, int, int)
    def update_reference_combobox_index_before_remove(self, _: QModelIndex, first: int,
                                                      last: int) -> None:
        previous_index = self.reference_combobox.currentIndex()
        if first <= previous_index <= last:
            # otherwise QComboBox set currentIndex to first-1
            self.reference_combobox.setCurrentIndex(0)

    @Slot(QModelIndex, int, int)
    def update_profile_combobox_index_before_remove(self, _: QModelIndex, first: int,
                                                    last: int) -> None:
        previous_index = self.profile_combobox.currentIndex()
        if first <= previous_index <= last:
            # otherwise QComboBox set currentIndex to first-1
            self.profile_combobox.setCurrentIndex(0)


# PointMarker class ################################################################################

class PointMarker(pg.InfiniteLine):
    """
    Based on pyqtgraph's InfiniteLine
    Vertical infinite line on plot, x-axis-position is synchronized with
    SpatialPlotView point slider position
    """

    pos_changed = Signal(int)

    base_pen = pg.functions.mkPen(0.5, width=3)
    dragged_pen = pg.functions.mkPen("cyan", width=3)
    hover_pen = pg.functions.mkPen("red", width=3)

    def __init__(self):
        self.length: Optional[int] = None
        super().__init__(pen=PointMarker.base_pen, hoverPen=PointMarker.hover_pen, angle=90,
                         movable=True)
        self.setZValue(-100)
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        self.sigDragged.connect(lambda _: self.setPen(PointMarker.dragged_pen))
        self.sigPositionChangeFinished.connect(lambda _: self.setPen(PointMarker.base_pen))

    def setPos(self, pos: Union[float, QPointF]) -> None:
        """
        Overloaded function
        Position can only be set to one of the values of the dataset
        (on x-axis, ie a band# or date)

        Parameters
        ----------
        pos : int, float, QPoint
            position (approx.) where the line should be set
        """
        if isinstance(pos, QPointF):
            pos = pos.x()
        if self.length is not None:
            pos, _ = get_nearest(range(self.length), pos)
        super().setPos(pos)

    def mouseDragEvent(self, ev) -> None:
        super().mouseDragEvent(ev)
        if ev.isAccepted():
            if ev.isStart():
                QGuiApplication.instance().setOverrideCursor(Qt.CursorShape.DragMoveCursor)
            elif ev.isFinish():
                QGuiApplication.instance().restoreOverrideCursor()



import logging

from PySide6.QtWidgets import (
    QTableView, QHeaderView
)
from PySide6.QtGui import (
    QColor, QStandardItemModel, QStandardItem
)
from PySide6.QtCore import (
    Qt, QObject, Signal, Slot,
    QAbstractItemModel, QModelIndex
)

class LogsStoreSignals(QObject):
    new_record = Signal()
class LogsStore(logging.Handler):
    def __init__(self):
        super().__init__()
        self._records = []
        self._signals = LogsStoreSignals()

    @property
    def records(self):
        return self._records
    @property
    def new_record(self):
        return self._signals.new_record

    def emit(self, record):
        self._records.append(record)
        self.new_record.emit()

class LogsItemModel(QAbstractItemModel):
    def __init__(self, log_store):
        super().__init__()
        self._log_store = log_store
        self._log_store.new_record.connect(self._on_new_record)

    @Slot()
    def _on_new_record(self):
        size = len(self._log_store.records)
        self.beginInsertRows(QModelIndex(), size-1, size-1)
        self.endInsertRows()

    def rowCount(self, index):
        if index.isValid():
            return 0
        else:
            return len(self._log_store.records)
    def columnCount(self, index):
        return 4

    def parent(self, index):
        return QModelIndex()

    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal:
            if role == Qt.ItemDataRole.DisplayRole:
                match section:
                    case 0:
                        return "Log level"
                    case 1:
                        return "Source"
                    case 2:
                        return "Function"
                    case 3:
                        return "Message"
        else:
            if role == Qt.ItemDataRole.DisplayRole:
                return str(section)
        return None

    def data(self, index, role):
        if index.isValid():
            record = self._log_store.records[index.row()]
            if role == Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return record.levelname
                    case 1:
                        return record.name
                    case 2:
                        return record.funcName
                    case 3:
                        return record.message

            if role == Qt.ItemDataRole.ForegroundRole:
                match record.levelno:
                    case logging.WARNING:
                        return QColor('orange')
                    case logging.ERROR:
                        return QColor('red')
                    case logging.DEBUG:
                        return QColor('grey')
                return None
            return None

    def index(self, row, column, parent):
        return self.createIndex(row, column, None)

class LoggerWidget(QTableView):
    def __init__(self, log_store):
        super().__init__()
        self.setModel(LogsItemModel(log_store))
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)



# -*- coding: utf-8 -*-

"""
Basic 3D vector operations.

Copyright (c) 2010, Renaud Blanch <rndblnch at gmail dot com>
Licence: GPLv3 or higher <http://www.gnu.org/licenses/gpl.html>
"""

# imports ####################################################################

from math import sqrt as _sqrt

_sum = sum


# globals ####################################################################

def _zip_zero_pad(*ps):
    len_max = max(len(p) for p in ps)
    ps = [tuple(p) + (0.,)*(len_max-len(p)) for p in ps]
    for t in zip(*ps):
        yield t


_O: tuple = tuple()
_3D = 3
_X, _Y, _Z = range(_3D)


# vector #####################################################################

def vector(p1=_O, p0=_O):
    return tuple(p1i-p0i for (p1i, p0i) in _zip_zero_pad(p1, p0))


def mul(a, v):
    return tuple(a*vi for vi in v)


def add(u, v):
    return tuple(ui+vi for (ui, vi) in _zip_zero_pad(u, v))


def sum(u=vector(), *vs):
    for v in vs:
        u = add(u, v)
    return u


def sub(u, v):
    return tuple(ui-vi for (ui, vi) in _zip_zero_pad(u, v))


def dot(u, v):
    return _sum(ui*vi for (ui, vi) in _zip_zero_pad(u, v))


def cross(u, v):
    assert len(u) == len(v) == _3D
    return (u[_Y]*v[_Z]-u[_Z]*v[_Y],
            u[_Z]*v[_X]-u[_X]*v[_Z],
            u[_X]*v[_Y]-u[_Y]*v[_X])


def norm(v):
    return _sqrt(dot(v, v))


def matrix(v):
    return [[float(vi)] for vi in v]



# -*- coding: utf-8 -*-

"""Bresenham algorithm for line drawing."""


def line(x0: int, y0: int, x1: int, y1: int) -> list[tuple[int, int]]:
    """returns list of indices of pixels for the line from (x0, y0) to (x1, y1)

    both ends are included.

    Parameters
    ----------
    x0 : int
        row of first point
    y0 : int
        col of first point
    x1 : int
        row of second point
    y1 : int
        col of second point

    Examples
    --------
    >>> from insarviz.bresenham import line
    >>> import numpy as np
    >>> X0, Y0, X1, Y1 = 0, 0, 4, 3
    >>> res = line(X0, Y0, X1, Y1)
    >>> print(res, type(res)==list, type(res[0])==tuple)
    [(0, 0), (1, 1), (2, 2), (3, 2), (4, 3)] True True
    """

    idxs: list[tuple[int, int]] = []
    dx, dy = x1-x0, y1-y0

    xsign = 1 if dx > 0 else -1
    ysign = 1 if dy > 0 else -1

    dx, dy = abs(dx), abs(dy)

    if dx > dy:
        xx, xy, yx, yy = xsign, 0, 0, ysign
    else:
        dx, dy = dy, dx
        xx, xy, yx, yy = 0, ysign, xsign, 0

    D = 2*dy - dx
    y = 0

    for x in range(dx + 1):
        idxs += [(x0 + x*xx + y*yx, y0 + x*xy + y*yy)]
        if D >= 0:
            y += 1
            D -= 2*dx
        D += 2*dy
    return idxs


# example:
# import numpy as np
# X0, Y0, X1, Y1 = 0, 0, 4, 3
# res = line(X0, Y0, X1, Y1)
# print(res, type(res)==list, type(res[0])==tuple)

# selmap = np.zeros((15,5))
# for i in res:
#     selmap[i] = 1.

# print('selmap: \n', selmap)


# X1, Y1, X2, Y2 = 4, 3, 8, 3
# res2 = line(X1, Y1, X2, Y2)
# print(res2)

# for i in res2:
#     selmap[i] = 1.

# print('selmap 2nd time: \n', selmap)

# X2, Y2, X3, Y3 = 8, 3, 7, 0
# res3 = line(X2, Y2, X3, Y3)
# print(res3)

# for i in res3:
#     selmap[i] = 1.

# print('selmap 3nd time: \n', selmap)



# -*- coding: utf-8 -*-

from typing import Any, Optional, TYPE_CHECKING

from PySide6.QtCore import Qt, QSize, Slot

from PySide6.QtGui import QIcon

from PySide6.QtWidgets import (
    QWidget, QMessageBox, QDialog, QDialogButtonBox, QGridLayout, QLineEdit, QLabel
)

import rasterio

import numpy as np

import pathlib

from owslib.wmts import WebMapTileService

from insarviz.map.layers.TileLayer import TileLayer

if TYPE_CHECKING:
    from insarviz.map.MapModel import MapModel
    from insarviz.map.AbstractMapView import AbstractMapView


# TEST WMTS servers

# self.wmts = WebMapTileService(r'http://tiles.maps.eox.at/wmts')
# self.layer = 'osm_3857'
# self.tilematrixset = 'GoogleMapsCompatible'

# self.wmts = WebMapTileService(r'https://data.geopf.fr/wmts')
# self.layer: str = 'ORTHOIMAGERY.ORTHOPHOTOS'
# self.tilematrixset: str = 'PM_0_21'

class WMTSLayer(TileLayer):

    icon: QIcon = QIcon()
    kind: str = "WMTS layer"

    def __init__(self, name: str, model: "MapModel", map_widget: "AbstractMapView",
                 server: str, layer: str, tilematrixset: str):
        super().__init__(name, model, map_widget)
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            WMTSLayer.icon = QIcon('icons:WMS.svg')
        try:
            self.wmts = WebMapTileService(server)
        except Exception as e:
            QMessageBox.critical(None, "Connection Error",
                                 f"Impossible to connect to {server}.\nIs internet available ?\n\n\nError:\n{repr(e)}")
            raise RuntimeError("Unable to connect to WMTS")
        if layer not in self.wmts.contents:
            QMessageBox.critical(None, "Invalid layer",
                                 f"Layer {layer} not found in {server}.")
            raise KeyError(f"Layer {self.layer} not found in {server}")
        self.layer: str = layer
        if tilematrixset not in self.wmts.contents[self.layer].tilematrixsetlinks.keys():
            QMessageBox.critical(None, "Invalid tilematrix set",
                                 f"Tilematrix set {tilematrixset} not found for layer {layer} in {server}.")
            raise KeyError(
                f"Tilematrix set {tilematrixset} not found for layer {layer} in {server}.")
        self.tilematrixset: str = tilematrixset
        self.format: str
        if "image/jpeg" in self.wmts.contents[self.layer].formats:
            self.format = "image/jpeg"
        elif "image/png" in self.wmts.contents[self.layer].formats:
            self.format = "image/png"
        else:
            QMessageBox.critical(None, "Invalid image format",
                                 f"Wmts does not have jpeg or png formats for the required layer: {self.wmts.contents[self.layer].formats}.")
            raise RuntimeError(
                f"wmts does not have jpeg or png formats: {self.wmts.contents[self.layer].formats}")
        self.crs = rasterio.CRS.from_string(self.wmts.tilematrixsets[self.tilematrixset].crs)
        # some wmts lack operation name, don't know why
        for i, op in enumerate(self.wmts.operations):
            if (not hasattr(op, 'name')):
                self.wmts.operations[i].name = ""
        # to make the computations, the crs unit shall be meters, see https://www.ogc.org/standard/wmts/
        if self.crs.units_factor[0] not in ("metre", "meter"):
            QMessageBox.critical(None, "Incompatible dataset crs",
                                 "In order to use a WMTS layer, the main data crs unit must be 'meter' or 'metre'.")
            raise RuntimeError(
                f"main data crs unit must be 'meter' or 'metre'")
        assert not (rasterio.crs.epsg_treats_as_latlong(self.crs))
        for zoom, tilematrix in self.wmts.tilematrixsets[self.tilematrixset].tilematrix.items():
            # compute the corners of the tilematrix in its crs coordinates
            # see https://www.ogc.org/standard/wmts/
            left, top = tilematrix.topleftcorner
            pixelspan = tilematrix.scaledenominator * 0.00028 * self.crs.units_factor[1]
            right = left + pixelspan * tilematrix.tilewidth * tilematrix.matrixwidth
            bottom = top - pixelspan * tilematrix.tileheight * tilematrix.matrixheight
            # transform into coordinates in the main dataset crs
            left, bottom, right, top = rasterio.warp.transform_bounds(self.crs, self.main_crs,
                                                                      left, bottom, right, top)
            # the antimeridian can have the same value for left and right in some crs
            if np.isclose(left, right):
                left, right = -np.abs(left), np.abs(right)
            # transform into coordinates in main dataset pixels
            left, top = ~self.main_transform * (left, top)
            right, bottom = ~self.main_transform * (right, bottom)
            tile_to_texture_pixelratioX = (right - left) / \
                (tilematrix.tilewidth * tilematrix.matrixwidth)
            tile_to_texture_pixelratioY = (bottom - top) / \
                (tilematrix.tileheight * tilematrix.matrixheight)
            self.tile_to_texture_pixelratio[int(zoom)] = tile_to_texture_pixelratioX
            # np.mean([tile_to_texture_pixelratioX,tile_to_texture_pixelratioY])

    def tilematrix_xy_to_tilematrix_crs(self, zoom: int, x: int, y: int) -> tuple[float, float, float, float]:
        """
        Return the rect of the tile in self.crs coordinates as left, bottom, right, top
        """
        try:
            tilematrix = self.wmts.tilematrixsets[self.tilematrixset].tilematrix[str(zoom)]
        except KeyError:
            raise ValueError(f'zoom {zoom} is not in tilematrixset {self.tilematrixset}')
        if x < 0 or x >= tilematrix.matrixwidth:
            raise ValueError(f"x {x} is < 0 or >= tilematrix.width")
        if y < 0 or y >= tilematrix.matrixheight:
            raise ValueError(f"y {y} is < 0 or >= tilematrix height")
        tilematrixlimits = self.wmts.contents[self.layer].tilematrixsetlinks[self.tilematrixset].tilematrixlimits
        if tilematrixlimits:
            tilematrixlimit = tilematrixlimits[str(zoom)]
            if x < tilematrixlimit.mintilecol or x > tilematrixlimit.maxtilecol:
                raise ValueError(f"x {x} is out of tilematrixlimit")
            if y < tilematrixlimit.mintilerow or y > tilematrixlimit.maxtilerow:
                raise ValueError(f"y {y} is out of tilematrixlimit")
        # compute the corner of the tile xy in its crs coordinates
        pixelspan = tilematrix.scaledenominator * 0.00028 * self.crs.units_factor[1]
        left, top = tilematrix.topleftcorner
        left = left + pixelspan * x * tilematrix.tilewidth
        top = top - pixelspan * y * tilematrix.tileheight
        right = left + pixelspan * tilematrix.tilewidth
        bottom = top - pixelspan * tilematrix.tileheight
        return left, bottom, right, top

    def request_url(self, zoom: int, x: int, y: int) -> str:
        """
        Return the url to request the tile as a string
        """
        request = self.wmts.buildTileRequest(layer=self.layer, tilematrixset=self.tilematrixset,
                                             tilematrix=str(zoom), row=y, column=x,
                                             format=self.format)
        return self.wmts.url + '?' + request

    def tile_shape(self, zoom: int) -> tuple[int, int]:
        """
        Return the shape (width, height) in pixels of a tile at the given zoom level
        """
        try:
            tilematrix = self.wmts.tilematrixsets[self.tilematrixset].tilematrix[str(zoom)]
        except KeyError:
            raise ValueError(f'zoom {zoom} is not in tilematrixset {self.tilematrixset}')
        return (tilematrix.tilewidth, tilematrix.tileheight)

    def main_crs_to_tilematrix_xy(self, left: float, bottom: float, right: float, top: float,
                                  zoom: int) -> tuple[int, int, int, int]:
        """
        Return the rect of tiles in xy coordinates as left, bottom, right, top
        """
        try:
            tilematrix = self.wmts.tilematrixsets[self.tilematrixset].tilematrix[str(zoom)]
        except KeyError:
            raise ValueError(f'zoom {zoom} is not in tilematrixset {self.tilematrixset}')
        # transform into coordinates in the tilematrix crs
        left, bottom, right, top = rasterio.warp.transform_bounds(
            self.main_crs, self.crs, left, bottom, right, top)
        tilematrix_bounds = self.tilematrix_bounds(zoom)
        # if both rects do not intersect return None
        if left > tilematrix_bounds[2]:
            return None
        if bottom > tilematrix_bounds[3]:
            return None
        if right < tilematrix_bounds[0]:
            return None
        if top < tilematrix_bounds[1]:
            return None
        # take the intersection between both rects
        left = max(left, tilematrix_bounds[0])
        bottom = max(bottom, tilematrix_bounds[1])
        right = min(right, tilematrix_bounds[2])
        top = min(top, tilematrix_bounds[3])
        # compute the x and y values
        pixelspan = tilematrix.scaledenominator * 0.00028 * self.crs.units_factor[1]
        left = int((left - tilematrix.topleftcorner[0]) / (pixelspan * tilematrix.tilewidth))
        bottom = int(
            (tilematrix.topleftcorner[1] - bottom) / (pixelspan * tilematrix.tileheight))
        right = int((right - tilematrix.topleftcorner[0]) / (pixelspan * tilematrix.tilewidth))
        top = int((tilematrix.topleftcorner[1] - top) / (pixelspan * tilematrix.tileheight))
        return left, bottom, right, top

    def tilematrix_bounds(self, zoom: int) -> tuple[float, float, float, float]:
        if str(zoom) in self.wmts.tilematrixsets[self.tilematrixset].tilematrix:
            tilematrix = self.wmts.tilematrixsets[self.tilematrixset].tilematrix[str(zoom)]
            pixelspan = tilematrix.scaledenominator * 0.00028 * self.crs.units_factor[1]
            left, top = tilematrix.topleftcorner
            tilematrixlimits = self.wmts.contents[self.layer].tilematrixsetlinks[self.tilematrixset].tilematrixlimits
            if tilematrixlimits:
                tilematrixlimit = tilematrixlimits[str(zoom)]
                left = left + pixelspan * tilematrixlimit.mintilecol * tilematrix.tilewidth
                top = top + pixelspan * tilematrixlimit.mintilerow * tilematrix.tileheight
                right = left + pixelspan * tilematrixlimit.maxtilecol * tilematrix.tilewidth
                bottom = top - pixelspan * tilematrixlimit.maxtilerow * tilematrix.tileheight
            else:
                right = left + pixelspan * tilematrix.matrixwidth * tilematrix.tilewidth
                bottom = top - pixelspan * tilematrix.matrixheight * tilematrix.tileheight
            return left, bottom, right, top
        else:
            raise ValueError(f'zoom {zoom} is not in tilematrixset {self.tilematrixset}')

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        output["server"] = self.wmts.url
        output["wmts_layer"] = self.layer
        output["tilematrixset"] = self.tilematrixset
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel",
                  map_widget: "AbstractMapView") -> "WMTSLayer":
        assert input_dict["kind"] == cls.kind
        name = input_dict.get("name", "WMTS layer")
        server = input_dict["server"]
        layer = input_dict["wmts_layer"]
        tilematrixset = input_dict["tilematrixset"]
        layer = WMTSLayer(name, map_model, map_widget, server, layer, tilematrixset)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        if "alpha" in input_dict:
            layer.alpha = input_dict["alpha"]
        return layer


class WMTSLayerDialog(QDialog):

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setModal(True)
        self.setWindowTitle("Create WMTS Layer")
        # labels
        self.server_label = QLabel("Server url:")
        self.layer_label = QLabel("Layer:")
        self.tilematrixset_label = QLabel("Tilematrix set:")
        # inputs
        self.server_input = QLineEdit()
        self.server_input.setPlaceholderText("http://...")
        self.server_input.textChanged.connect(self.test_enable_ok)
        self.layer_input = QLineEdit()
        self.layer_input.setPlaceholderText("Layer name")
        self.layer_input.textChanged.connect(self.test_enable_ok)
        self.tilematrixset_input = QLineEdit()
        self.tilematrixset_input.setPlaceholderText("Tilematrix set name")
        self.tilematrixset_input.textChanged.connect(self.test_enable_ok)
        # buttons
        self.button_box = QDialogButtonBox()
        self.cancel_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.cancel_button.clicked.connect(self.reject)
        self.ok_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.ok_button.clicked.connect(self.accept)
        # main layout
        self.main_layout = QGridLayout()
        self.main_layout.addWidget(self.server_label, 0, 0)
        self.main_layout.addWidget(self.server_input, 0, 1)
        self.main_layout.addWidget(self.layer_label, 1, 0)
        self.main_layout.addWidget(self.layer_input, 1, 1)
        self.main_layout.addWidget(self.tilematrixset_label, 2, 0)
        self.main_layout.addWidget(self.tilematrixset_input, 2, 1)
        self.main_layout.addWidget(self.button_box, 3, 0, 1, 2)
        self.setLayout(self.main_layout)
        self.test_enable_ok()

    @Slot()
    def test_enable_ok(self) -> None:
        enable_ok = True
        if self.server_input.text() == "":
            enable_ok = False
        if self.layer_input.text() == "":
            enable_ok = False
        if self.tilematrixset_input.text() == "":
            enable_ok = False
        self.ok_button.setEnabled(enable_ok)

    def sizeHint(self) -> QSize:
        return QSize(512, super().sizeHint().height())



from typing import Optional, Any
import pathlib

from PySide6.QtGui import QPainter, QIcon
from PySide6.QtOpenGL import QOpenGLVertexArrayObject, QAbstractOpenGLFunctions

from .Layer import Layer
from ..TreeItemAttribute import TreeItemSliderAttribute, TreeItemFloatAttribute
from ...linalg import matrix
from ...Roles import Roles

class SwipeTool(Layer):
    kind = "swipe layer"
    icon: QIcon = QIcon()

    def __init__(self, name="Swipe tool", cutoff=0, dataset=None):
        super().__init__(name)
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            SwipeTool.icon = QIcon("icons:swipetool.png")
        self.cutoff = cutoff
        self.dataset = dataset
        self.add_child(TreeItemSliderAttribute(self,
                                               "cutoff",
                                               tooltip="cutoff value",
                                               vmin=0.0,
                                               vmax=1.0))

    def data(self, column: int, role: int):
        if role == Roles.DatasetRole:
            return self.dataset
        return super().data(column, role)

    def update_show_params(self, params: Layer.ShowParams):
        params.cutoff = max(self.cutoff, params.cutoff)

    def show(self, view_matrix: matrix.Matrix, projection_matrix: matrix.Matrix, show_params: Layer.ShowParams,
             painter: Optional[QPainter] = None, vao: Optional[QOpenGLVertexArrayObject] = None,
             glfunc: Optional[QAbstractOpenGLFunctions] = None, blend: bool = True) -> None:
        """
        Swipe layers aren't shown as such, they only affect the layers above them
        """
        return None

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        output["cutoff"] = self.cutoff
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel") -> "SwipeTool":
        assert input_dict["kind"] == cls.kind
        assert "cutoff" in input_dict

        name = input_dict.get("name", "Swipe tool")
        cutoff = input_dict["cutoff"]
        layer = SwipeTool(name, cutoff, dataset = map_model.loader.dataset)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        return layer



# -*- coding: utf-8 -*-
"""
LayerView

This module contains the LayerView and LayerWidget classes managing the layer manager widget.
"""

# imports ##########################################################################################

from typing import Optional
import rasterio

from PySide6.QtCore import (
    Qt, QObject, Signal, Slot, QModelIndex, QSize, QItemSelectionModel,
    QSortFilterProxyModel, QAbstractItemModel, QEvent
)

from PySide6.QtGui import QIcon, QPalette, QPainter, QImage, QPen, QBrush, QColor

from PySide6.QtWidgets import (
    QSlider,
    QTreeView, QAbstractItemView, QStyledItemDelegate,
    QHeaderView, QApplication, QStyle, QStyleOptionButton, QStyleOptionViewItem,
    QSpinBox, QDoubleSpinBox, QColorDialog, QPushButton, QWidget, QDialog, QLabel,
    QVBoxLayout, QLayout, QSizePolicy, QWidget
)

from pyqtgraph.colormap import ColorMap

from insarviz.map.layers.LayerModel import LayerModel, SetDataCommand

from insarviz.map.layers.RasterLayerColorMapEditor import RasterLayerColorMapEditor

from insarviz.Roles import Roles

class SliderWithValue(QWidget):
    SLIDER_SENSITIVITY: float = 1000.0

    sliderMoved = Signal(float)
    sliderReleased = Signal()

    def __init__(self, parent = None, display_value = str, *args):
        super().__init__(parent)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        self.display_value = display_value
        self.initial_value = None

        self.slider = QSlider()
        self.slider.setFocusProxy(self)

        label = QLabel("0")
        label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)

        def set_label_value(v):
            label.setText(self.display_value(v / self.SLIDER_SENSITIVITY))
        self.slider.valueChanged.connect(set_label_value)
        def on_slider_moved(v):
            self.sliderMoved.emit(v / self.SLIDER_SENSITIVITY)
        self.slider.sliderMoved.connect(on_slider_moved)
        self.slider.sliderMoved.connect(set_label_value)
        self.slider.sliderReleased.connect(lambda: self.sliderReleased.emit())
        def on_slider_pressed():
            self.initial_value = self.value()
        self.slider.sliderPressed.connect(on_slider_pressed)

        layout = QVBoxLayout(self)
        layout.setSizeConstraint(QLayout.SizeConstraint.SetMinimumSize)
        layout.addWidget(self.slider)
        layout.addWidget(label)

    def setMinimum(self, vmin):
        self.slider.setMinimum(int(vmin * self.SLIDER_SENSITIVITY))
    def setMaximum(self, vmax):
        self.slider.setMaximum(int(vmax * self.SLIDER_SENSITIVITY))
    def setValue(self, value):
        self.slider.setValue(int(value * self.SLIDER_SENSITIVITY))
    def value(self):
        return self.slider.value() / self.SLIDER_SENSITIVITY

    def setDisplayValue(self, display_value):
        self.display_value = display_value


# LayerView class ##################################################################################
class LayerView(QTreeView):

    # TODO display branches ? https://doc.qt.io/qt-6/stylesheet-examples.html#customizing-qtreeview

    def __init__(self, layer_model: LayerModel):
        super().__init__()
        self.layer_model = layer_model
        # use proxy model to remove empty SelectionFolders
        self.proxy_model = ProxyLayerModel()
        self.proxy_model.setSourceModel(self.layer_model)
        self.setModel(self.proxy_model)
        # formatting of the view
        self.setHeaderHidden(True)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        palette: QPalette = QPalette(self.palette())
        palette.setColor(QPalette.ColorRole.Highlight, Qt.GlobalColor.white)
        palette.setColor(QPalette.ColorRole.HighlightedText, Qt.GlobalColor.black)
        self.setPalette(palette)
        # Not all rows have the same height (slider layers)
        # self.setUniformRowHeights(True)
        self.setItemDelegateForColumn(0, ItemDelegate(self))
        self.remove_delegate = RemoveDelegate(self)
        self.setItemDelegateForColumn(1, self.remove_delegate)
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        self.header().setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.showDropIndicator()
        self.setMouseTracking(True)  # needed for self.entered signal
        # signals and slots connection
        self.entered.connect(self.create_remove_editor)
        self.remove_delegate.clicked.connect(self.remove)
        self.selectionModel().currentChanged.connect(
            lambda x, _: self.layer_model.manage_current_changed(self.proxy_model.mapToSource(x)))
        self.layer_model.rowsMoved.connect(self.update_current_from_move)
        self.layer_model.request_expand.connect(
            lambda i: self.expand(self.proxy_model.mapFromSource(i)))

    # connectd to LayerView.entered
    @Slot(QModelIndex)
    def create_remove_editor(self, index: QModelIndex) -> bool:
        if not index.isValid():
            return False
        if index.column() == self.layer_model.remove_column and \
                self.selectionModel().isRowSelected(index.row(), index.parent()):
            self.edit(index)
            return True
        return False

    # connected to RemoveDelegate.clicked
    @Slot(QModelIndex)
    def remove(self, index: QModelIndex) -> None:
        if index.isValid():
            target: QModelIndex = self.proxy_model.mapToSource(index.siblingAtColumn(0))
            self.layer_model.remove(target)

    # connected to LayerModel.rowsMoved
    @Slot(QModelIndex, int, int, QModelIndex, int)
    def update_current_from_move(self, parent: QModelIndex, start: int, end: int, dest: QModelIndex,
                                 row: int) -> None:
        """
        Update self.currentIndex when layers are moved (drag and drop or up/down button)
        """
        parent = self.proxy_model.mapFromSource(parent)
        dest = self.proxy_model.mapFromSource(dest)
        if parent == QModelIndex() and parent == dest and start == end:
            flags = (QItemSelectionModel.SelectionFlag.SelectCurrent |
                     QItemSelectionModel.SelectionFlag.Rows)
            if start < row:
                row -= 1
            index: QModelIndex = self.proxy_model.index(row, 0, parent)
            self.selectionModel().clear()
            self.selectionModel().setCurrentIndex(index, flags)


# ProxyLayerModel class ############################################################################

class ProxyLayerModel(QSortFilterProxyModel):
    """
    Proxy filter model that filters out empty selection folders. As recursive filtering is enabled,
    the filter is applied on parents (i.e. selection folders) when their children (i.e. selection
    items) are added/modified/removed.
    """

    def __init__(self):
        super().__init__()
        self.setRecursiveFilteringEnabled(True)

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        # pylint: disable=missing-function-docstring, invalid-name
        index: QModelIndex = self.sourceModel().index(source_row, 0, source_parent)
        if self.sourceModel().data(index, Roles.FolderRole) == "SELECTION_FOLDER":
            if self.sourceModel().rowCount(index) == 0:
                # if empty SelectionFolder then filter out
                return False
        return True


# ItemDelegate class ###############################################################################

class ItemDelegate(QStyledItemDelegate):
    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)

    def createEditor(self, parent: QWidget, option: QStyleOptionViewItem,
                     index: QModelIndex) -> QWidget:
        editor_role: Optional[str] = index.data(Roles.EditorRole)
        editor: QWidget
        if editor_role == "integer":
            editor = QSpinBox(parent)
            editor.setFrame(False)
            return editor
        if editor_role == "float":
            editor = QDoubleSpinBox(parent)
            editor.setFrame(False)
            editor.setDecimals(2)
            editor.setSingleStep(0.1)
            return editor
        if editor_role == "float_slider":
            editor = SliderWithValue(parent)
            editor.slider.setOrientation(Qt.Horizontal)
            editor.model_index = index
            def set_model_data(v):
                editor.model_index.model().setData(editor.model_index, v)
            editor.sliderMoved.connect(set_model_data)
            index.model().setData(index, editor, Roles.PersistentEditorWidgetRole)
            editor.sliderReleased.connect(lambda: self.commitData.emit(editor))
            return editor
        if editor_role == "color":
            editor = QColorDialog(parent)
            editor.setModal(True)
            return editor
        if editor_role == "colormap":
            editor = RasterLayerColorMapEditor(parent)
            return editor
        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: QWidget, index: QModelIndex) -> None:
        editor_role: Optional[str] = index.data(Roles.EditorRole)
        if editor_role == "integer":
            assert isinstance(editor, QSpinBox)
            value, name, vmin, vmax, unit = index.data(Qt.ItemDataRole.EditRole)
            editor.setValue(value)
            editor.setPrefix(f"{name} : ")
            if vmin is not None:
                editor.setMinimum(vmin)
            if vmax is not None:
                editor.setMaximum(vmax)
            if unit:
                editor.setSuffix(f" {unit}")
        elif editor_role == "float":
            assert isinstance(editor, QDoubleSpinBox)
            value, name, vmin, vmax, unit = index.data(Qt.ItemDataRole.EditRole)
            editor.setValue(value)
            editor.setPrefix(f"{name} : ")
            if vmin is not None:
                editor.setMinimum(vmin)
            if vmax is not None:
                editor.setMaximum(vmax)
            if unit:
                editor.setSuffix(f" {unit}")
        elif editor_role == "float_slider":
            value, name, vmin, vmax = index.data(Qt.ItemDataRole.EditRole)
            if vmin is not None:
                editor.setMinimum(vmin)
            if vmax is not None:
                editor.setMaximum(vmax)
            dataset = index.model().data(index.parent(), Roles.DatasetRole)
            def show_lat(v):
                x_d, y_d = dataset.xy(0.0, dataset.width * v)
                if dataset.crs is not None:
                    [x], [y] = rasterio.warp.transform(dataset.crs, rasterio.crs.CRS.from_epsg(4326),
                                                       [x_d], [y_d])
                else:
                    x, y = x_d, y_d
                return f"long: {x:0.03f}"
            editor.setDisplayValue(show_lat)
            editor.setValue(value)
        elif editor_role == "color":
            assert isinstance(editor, QColorDialog)
            color = index.data(Qt.ItemDataRole.EditRole)
            editor.setCurrentColor(color)
        elif editor_role == "colormap":
            assert isinstance(editor, RasterLayerColorMapEditor)
            layer = index.data(Qt.ItemDataRole.EditRole)
            editor.set_layer(layer)
        else:
            super().setEditorData(editor, index)

    def setModelData(self, editor: QWidget, model: QAbstractItemModel, index: QModelIndex) -> None:
        assert isinstance(model, ProxyLayerModel)
        index = model.mapToSource(index)
        model = model.sourceModel()
        assert isinstance(model, LayerModel)
        editor_role: Optional[str] = index.data(Roles.EditorRole)
        if editor_role == "integer":
            assert isinstance(editor, QSpinBox)
            editor.interpretText()
            model.add_undo_command.emit(SetDataCommand(
                model, index, int(editor.value()), Qt.ItemDataRole.EditRole))
        elif editor_role == "float":
            assert isinstance(editor, QDoubleSpinBox)
            editor.interpretText()
            model.add_undo_command.emit(SetDataCommand(
                model, index, float(editor.value()), Qt.ItemDataRole.EditRole))
        elif editor_role == "float_slider":
            model.add_undo_command.emit(SetDataCommand(
                model, index, float(editor.value()), Qt.ItemDataRole.EditRole,
                old_value = editor.initial_value
            ))
        elif editor_role == "color":
            assert isinstance(editor, QColorDialog)
            if editor.result() == QDialog.DialogCode.Accepted:
                color = editor.currentColor()
                model.add_undo_command.emit(SetDataCommand(model, index, color,
                                                           Qt.ItemDataRole.EditRole))
        elif editor_role == "colormap":
            assert isinstance(editor, RasterLayerColorMapEditor)
            if editor.result() == QDialog.DialogCode.Accepted:
                colormap = editor.get_colormap()
                v0, v1 = editor.get_v0_v1()
                model.add_undo_command.emit(SetDataCommand(model, index, (colormap, v0, v1),
                                                           Qt.ItemDataRole.EditRole))
        else:
            # base implementation of QStyledItemDelegate with model.setData replaced by
            # model.add_undo_command...
            n = editor.metaObject().userProperty().name()
            if n == "":
                n = str(self.itemEditorFactory().valuePropertyName(
                    model.data(index, Qt.ItemDataRole.EditRole).userType()))
            if n != "":
                model.add_undo_command.emit(SetDataCommand(
                    model, index, editor.property(n), Qt.ItemDataRole.EditRole))

    def updateEditorGeometry(self, editor: QWidget, option: QStyleOptionViewItem,
                             index: QModelIndex) -> None:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        editor.setGeometry(option.rect)

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex):
        persistent_editor = index.model().data(index, int(Roles.PersistentEditorWidgetRole))
        if persistent_editor is not None:
            ret = persistent_editor.sizeHint()
            return ret
        return super().sizeHint(option, index)

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex):
        display_role = index.data(Qt.ItemDataRole.DisplayRole)
        if display_role is None:
            persistent_editor = index.data(int(Roles.PersistentEditorWidgetRole))
            if persistent_editor is None:
                self.parent().openPersistentEditor(index)
            else:
                persistent_editor.model_index = index

        elif isinstance(display_role, QColor):
            self.initStyleOption(option, index)
            painter.save()
            color: QColor = index.data(Qt.ItemDataRole.DisplayRole)
            img = QImage(1, 1, QImage.Format.Format_RGBA8888)
            img.setPixelColor(0, 0, color)
            painter.drawImage(option.rect, img)
            painter.setBrush(QBrush())
            painter.setPen(QPen(Qt.GlobalColor.black, 1, Qt.PenStyle.SolidLine))
            painter.drawRect(option.rect.adjusted(0, 0, 0, -1))
            painter.restore()
        elif isinstance(display_role, ColorMap):
            self.initStyleOption(option, index)
            painter.save()
            colormap: ColorMap = index.data(Qt.ItemDataRole.DisplayRole)
            lut = colormap.getLookupTable(nPts=option.rect.width(), alpha=True)
            img = QImage(lut, len(lut), 1, QImage.Format.Format_RGBA8888)
            painter.drawImage(option.rect, img)
            painter.setBrush(QBrush())
            painter.setPen(QPen(Qt.GlobalColor.black, 1, Qt.PenStyle.SolidLine))
            painter.drawRect(option.rect.adjusted(0, 0, 0, -1))
            painter.restore()

        else:
            super().paint(painter, option, index)

    def eventFilter(self, editor: QObject, event: QEvent) -> bool:
        if isinstance(editor, (QColorDialog, RasterLayerColorMapEditor)):
            # tab is not used to move to the next element when editor is opened
            if event.type() == QEvent.Type.KeyPress:
                if event.key() == Qt.Key.Key_Tab or event.key() == Qt.Key.Key_Backtab:
                    return False
                elif (event.key() == Qt.Key.Key_Enter or event.key() == Qt.Key.Key_Return or
                      event.key() == Qt.Key.Key_Escape):
                    return False
            # prevent editor to close when loosing focus (resize / alt+tab...)
            if event.type() == QEvent.Type.FocusOut:
                return False
        return super().eventFilter(editor, event)


# RemoveDelegate class #############################################################################

class RemoveDelegate(QStyledItemDelegate):

    clicked = Signal(QModelIndex)

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        self.remove_icon = QIcon('icons:remove.png')

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex) -> None:
        if not index.isValid():
            return None
        self.initStyleOption(option, index)
        if not option.state & QStyle.StateFlag.State_Selected:
            return None
        if not index.flags() & Qt.ItemFlag.ItemIsEditable:
            # if item is not deletable
            return None
        if option.widget:
            style = option.widget.style()
        else:
            app = QApplication.instance()
            assert isinstance(app, QApplication)
            style = app.style()
        widget = option.widget
        style.drawControl(QStyle.ControlElement.CE_PushButton, self.button_option(option), painter,
                          widget)

    def createEditor(self, parent: QWidget, option: QStyleOptionViewItem,
                     index: QModelIndex) -> QWidget:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        editor = QPushButton(parent)
        editor.setFlat(True)
        editor.setIcon(self.remove_icon)
        editor.setToolTip("Remove")
        editor.clicked.connect(lambda: self.clicked.emit(index))
        return editor

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex) -> QSize:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        # iconsize + 8 pixels vertically (4 above and 4 under)
        button_option = self.button_option(option)
        return button_option.iconSize + QSize(0, 8)

    def updateEditorGeometry(self, editor: QWidget, option: QStyleOptionViewItem,
                             index: QModelIndex) -> None:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        editor.setGeometry(option.rect)

    def button_option(self, option: QStyleOptionViewItem) -> QStyleOptionButton:
        # see https://forum.qt.io/topic/131602/set-delegate-for-each-cell-in-a-qtablewidget/21
        button_option = QStyleOptionButton()
        button_option.rect = option.rect
        button_option.icon = self.remove_icon
        if option.widget:
            style = option.widget.style()
        else:
            app = QApplication.instance()
            assert isinstance(app, QApplication)
            style = app.style()
        size = style.pixelMetric(QStyle.PixelMetric.PM_ButtonIconSize, option)
        button_option.iconSize = QSize(size, size)
        button_option.palette = option.palette
        button_option.features = (QStyleOptionButton.ButtonFeature.None_ |
                                  QStyleOptionButton.ButtonFeature.Flat)
        button_option.state = option.state
        return button_option



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""MapModel

This module contains the MapModel class that manages data for the Map and
Minimap views (as in a Model/View architecture).
"""

# imports ###################################################################

from typing import Any, Optional, TYPE_CHECKING

import logging

import pathlib

from PySide6.QtCore import QObject, Signal, Slot, QModelIndex

from PySide6.QtGui import QOpenGLContext, QOffscreenSurface

from PySide6.QtWidgets import QMessageBox

from PySide6.QtOpenGL import QOpenGLTexture

from OpenGL import GL

from OpenGL.constant import IntConstant

import numpy as np

from pyqtgraph.colormap import ColorMap

from insarviz.utils import normalize_path

from insarviz.colormaps import create_colormap_texture

from insarviz.Loader import Loader

from insarviz.map.Shaders import DATA_UNIT, COLORMAP_UNIT

from insarviz.map.layers.LayerModel import LayerModel

from insarviz.map.layers.Layer import Layer, MainLayer, Raster1BLayer, RasterRGBLayer

from insarviz.map.layers.WMTSLayer import WMTSLayer

from insarviz.map.layers.XYZLayer import XYZLayer, OpenStreetMapLayer

from insarviz.map.layers.SelectionLayer import SelectionLayer

from insarviz.map.layers.SwipeTool import SwipeTool

if TYPE_CHECKING:
    from insarviz.map.AbstractMapView import AbstractMapView

import insarviz.version as version

logger = logging.getLogger(__name__)


# Map Model class ##################################################################################

class MapModel(QObject):
    """
    Model managing data for Map and Minimap
    """
    closed = Signal()
    opened = Signal()

    # connected to MainLayer.set_texture
    # (texture_unit, texture)
    texture_changed = Signal(IntConstant, QOpenGLTexture)

    # connected to MainLayer.set_v0_v1, ColomapWidget.setLevels
    v0_v1_changed = Signal(float, float)  # (v0, v1)

    #  connected to ColormapWidget.set_colormap
    request_colormap = Signal(str)

    # connected to Loader.compute_histograms
    request_histograms = Signal(tuple)

    # connected to ColormapWidget.set_XXX_histogram
    total_hist_changed = Signal(np.ndarray, np.ndarray)  # (values, bins)
    band_hist_changed = Signal(np.ndarray, np.ndarray)  # (values, bins)

    # connected to AbstractMapView.paint
    request_paint = Signal()

    # connected to AbstractMapView.set_view_center
    request_set_view_center = Signal(float, float)  # (cx, cy)

    def __init__(self, loader: Loader, context: QOpenGLContext,
                 offscreen_surface: QOffscreenSurface):
        """
        MapModel
        """
        super().__init__()
        self.loader = loader
        self.request_histograms.connect(self.loader.compute_histograms)
        self.loader.data_loaded.connect(self.on_data_loaded)
        self.loader.histograms_computed.connect(self.on_histograms_computed)
        self.loader.data_invalidated.connect(self.on_data_invalidated)
        self.context = context  # QOpenGLContext shared with every other contexts
        self.offscreen_surface = offscreen_surface  # QOffscreenSurface to use OpenGL commands
        self.layer_model = LayerModel()
        self.layer_model.request_paint.connect(self.request_paint)
        self.textures: dict[int, int] = {}  # associate a band index to an openGL texture id
        self.histograms: dict[int, tuple[np.ndarray, np.ndarray]] = {}
        self.flip_h: bool = False  # image vertical flip state
        self.flip_v: bool = False  # image vertical flip state
        # current band index (starting from 1 like in gdal)
        self.current_band_index: Optional[int] = None
        self.colormap: Optional[ColorMap] = None
        # OpenGL texture for colormap
        self.colormap_texture = QOpenGLTexture(QOpenGLTexture.Target.Target1D)
        self.tex_width: int = 0  # data width (in pixels)
        self.tex_height: int = 0  # data height (in pixels)
        self.colormap_v0: float = 0.  # minimum value for colorbar range (default 5 percentile)
        self.colormap_v1: float = 1.  # maximum value for colorbar range (default 95 percentile)

    @Slot(int)
    def on_data_loaded(self, band_index: int) -> None:
        assert self.loader.dataset is not None
        self.tex_width, self.tex_height = self.loader.dataset.width, self.loader.dataset.height
        assert band_index in self.loader.dataset.indexes
        self.current_band_index = band_index
        self.request_set_view_center.emit(self.tex_width/2, self.tex_height/2)

    @Slot()
    def on_histograms_computed(self) -> None:
        assert self.loader.total_histogram is not None
        self.total_hist_changed.emit(*self.loader.total_histogram)
        if self.current_band_index is not None:
            self.band_hist_changed.emit(*self.loader.band_histograms[self.current_band_index-1])

    @Slot()
    def on_data_invalidated(self) -> None:
        # delete the OpenGL textures
        self.context.makeCurrent(self.offscreen_surface)
        for texture in self.textures.values():
            texture.destroy()
        self.context.doneCurrent()
        self.textures = {}
        if self.current_band_index is not None:
            self.show_band(self.current_band_index)

    def close(self) -> None:
        self.loader.close()
        # delete the OpenGL textures
        self.context.makeCurrent(self.offscreen_surface)
        for texture in self.textures.values():
            texture.destroy()
        self.colormap_texture.destroy()
        self.context.doneCurrent()
        self.textures = {}
        self.histograms = {}
        self.flip_h = False
        self.flip_v = False
        self.current_band_index = None
        self.tex_width = 0
        self.tex_height = 0
        self.colormap = None
        self.colormap_v0 = 0.
        self.colormap_v1 = 1.
        self.closed.emit()
        self.layer_model.clear()
        self.total_hist_changed.emit(np.array([]), np.array([]))
        self.band_hist_changed.emit(np.array([]), np.array([]))
        self.request_paint.emit()

    def create_base_layers(self) -> None:
        # emiting the signal before displaying the message box result in an error as the envent loop
        # does not launch the asyncio coroutine connected to the signal (don't know why)
        QMessageBox.information(None, "Computing histograms",
                                "The Colormap Panel is disabled until histograms are done.")
        self.request_histograms.emit(None)
        # initialize main_layer
        main_layer = MainLayer(self)
        self.layer_model.add_layer(main_layer)
        # initialize selection and selection_layer
        selection = SelectionLayer()
        self.layer_model.add_layer(selection, i=0)
        self.layer_model.set_selection(selection)
        self.request_colormap.emit("greyscale")
        self.v0_v1_changed.emit(self.colormap_v0, self.colormap_v1)
        self.show_band(self.current_band_index)

    @Slot(int)
    def show_band(self, i: int) -> None:
        """
        Load, generate (if not existing) and show the texture of the ith band.

        Parameters
        ----------
        i : int
            Band/date number to be loaded and shown.

        Returns
        -------
        None.

        """
        assert self.loader.dataset is not None
        assert i in self.loader.dataset.indexes
        self.current_band_index = i
        # band data
        try:  # looking up cache
            self.textures[i]
        except KeyError:
            band = self.loader.load_band(i)
            band = band.astype(np.float32)
            # band data
            bg = np.isnan(band)
            h, w = band.shape
            # FIX for issue #183 (texture bug on mac M1/M2/M3)
            z = np.ones((h, w, 3), dtype='float32')
            z[:, :, 0] = band
            z[:, :, 0:2][bg] = 0.
            # band texture
            self.context.makeCurrent(self.offscreen_surface)
            texture = QOpenGLTexture(QOpenGLTexture.Target.Target2D)
            texture.setMagnificationFilter(QOpenGLTexture.Filter.Nearest)
            texture.setMinificationFilter(QOpenGLTexture.Filter.LinearMipMapLinear)
            texture.setWrapMode(QOpenGLTexture.WrapMode.ClampToEdge)
            texture.setSize(w, h)
            texture.setFormat(QOpenGLTexture.TextureFormat.RGB32F)
            texture.allocateStorage(QOpenGLTexture.PixelFormat.RGB,
                                    QOpenGLTexture.PixelType.Float32)
            texture.setData(QOpenGLTexture.PixelFormat.RGB,
                            QOpenGLTexture.PixelType.Float32, z.data)
            texture.generateMipMaps()
            self.context.doneCurrent()
            assert texture.textureId(
            ), f"Map Model : cannot load image texture for band {i} in OpenGl"
            # store band texture param
            self.textures[i] = texture
        # send the arrays of the histogram (values and bins)
        self.band_hist_changed.emit(*self.loader.band_histograms[self.current_band_index - 1])
        self.texture_changed.emit(IntConstant("GL_TEXTURE0+DATA_UNIT", GL.GL_TEXTURE0+DATA_UNIT),
                                  self.textures[i])
        self.request_paint.emit()

    # connected to ColormapWidget.colormap_changed, called when user change color gradient
    @Slot(ColorMap)
    def set_colormap(self, colormap: ColorMap) -> None:
        if self.loader.dataset is not None:
            self.colormap = colormap
            self.context.makeCurrent(self.offscreen_surface)
            self.colormap_texture.destroy()
            self.colormap_texture = create_colormap_texture(colormap)
            self.context.doneCurrent()
            self.texture_changed.emit(IntConstant("GL_TEXTURE0+COLORMAP_UNIT", GL.GL_TEXTURE0+COLORMAP_UNIT),
                                      self.colormap_texture)
            self.request_paint.emit()

    # connected to ColormapWidget.v0_v1_changed
    @Slot(float, float)
    def set_v0_v1(self, v0: float, v1: float) -> None:
        # prevents infinite looping between MapModel and ColormapWidgetView
        if self.colormap_v0 != v0 or self.colormap_v1 != v1:
            self.colormap_v0 = v0
            self.colormap_v1 = v1
            self.v0_v1_changed.emit(v0, v1)
            self.request_paint.emit()

    # connected to MainWindow.flip_h_action
    @Slot(bool)
    def set_flip_h(self, checked: bool) -> None:
        self.flip_h = checked
        self.request_paint.emit()

    # connected to MainWindow.flip_v_action
    @Slot(bool)
    def set_flip_v(self, checked: bool) -> None:
        self.flip_v = checked
        self.request_paint.emit()

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = {}
        output["insarviz"] = version.__version__
        try:
            output["dataset_path"] = self.loader.filepath.relative_to(
                project_path.parent).as_posix()
        except ValueError:
            output["dataset_path"] = self.loader.filepath.as_posix()
        output["layers"] = [layer.to_dict(project_path) for layer in self.layer_model.layers]
        output["flip_h"] = self.flip_h
        output["flip_v"] = self.flip_v
        output["current_band"] = self.current_band_index
        output["reference_band_index"] = self.loader.reference_band_index
        assert isinstance(self.colormap, ColorMap)
        output["colormap"] = self.colormap.name
        output["colormap_v0"] = self.colormap_v0
        output["colormap_v1"] = self.colormap_v1
        # numpy arrays need to be transformed into list in order to be serialized in json
        assert self.loader.total_histogram is not None
        output["total_histogram"] = tuple([a.tolist() for a in self.loader.total_histogram])
        assert self.loader.band_histograms is not None
        output["band_histograms"] = [tuple([a.tolist() for a in b])
                                     for b in self.loader.band_histograms]
        return output

    def from_dict(self, input_dict: dict[str, Any], project_path: str, map_widget: "AbstractMapView") -> bool:
        assert self.loader.dataset is not None
        if "reference_band_index" in input_dict:
            if input_dict["reference_band_index"] is not None and input_dict["reference_band_index"] not in self.loader.dataset.indexes:
                logger.warning("invalid reference_band_index")
                return False
        self.loader.set_reference_band(input_dict.get("reference_band_index", None))
        selection_layer: Optional[SelectionLayer] = None
        main_layer: Optional[MainLayer] = None
        for layer_dict in input_dict["layers"]:
            layer: Layer
            if layer_dict["kind"] == SelectionLayer.kind:
                if selection_layer is not None:
                    logger.warning("More than one selection layer")
                    self.close()
                    return False
                selection_layer = SelectionLayer.from_dict(layer_dict)
                layer = selection_layer
            elif layer_dict["kind"] == MainLayer.kind:
                if main_layer is not None:
                    logger.warning("More than one main data layer")
                    self.close()
                    return False
                main_layer = MainLayer.from_dict(layer_dict, self)
                layer = main_layer
            elif layer_dict["kind"] == OpenStreetMapLayer.kind:
                layer = OpenStreetMapLayer.from_dict(layer_dict, self, map_widget)
            elif layer_dict["kind"] == "WMS Layer":
                # for retrocompatibility purpose (old name of openstreetmap layer)
                layer = OpenStreetMapLayer.from_dict(layer_dict, self, map_widget)
            elif layer_dict["kind"] == WMTSLayer.kind:
                layer = WMTSLayer.from_dict(layer_dict, self, map_widget)
            elif layer_dict["kind"] == XYZLayer.kind:
                layer = XYZLayer.from_dict(layer_dict, self, map_widget)
            elif layer_dict["kind"] == SwipeTool.kind:
                layer = SwipeTool.from_dict(layer_dict, self)
            elif layer_dict["kind"] == Raster1BLayer.kind:
                if "filepath" not in layer_dict:
                    logger.warning(f'Layer {layer_dict["kind"]} "{layer_dict["name"]}"'
                                   ' miss the filepath field, skipping that layer')
                    continue
                layer_dict["filepath"] = normalize_path(layer_dict.get("filepath"), project_path)
                layer = Raster1BLayer.from_dict(layer_dict, self)
            elif layer_dict["kind"] == RasterRGBLayer.kind:
                if "filepath" not in layer_dict:
                    logger.warning(f'Layer {layer_dict["kind"]} "{layer_dict["name"]}"'
                                   ' miss the filepath field, skipping that layer')
                    continue
                normalize_path(layer_dict.get("filepath"), project_path)
                layer = RasterRGBLayer.from_dict(layer_dict, self)
            else:
                logger.warning(f'Unrecognized layer type {layer_dict["kind"]} '
                               f'for layer "{layer_dict["name"]}", skipping that layer')
                continue
            self.layer_model.add_layer(layer)
        if main_layer is None:
            logger.warning("Missing main data layer")
            self.close()
            return False
        if selection_layer is None:
            logger.warning("Missing selection layer")
            self.close()
            return False
        self.layer_model.set_selection(selection_layer)
        # send signals to fake the addition of the SelectionItems of the input_dict
        selection_index = self.layer_model.index(selection_layer.child_number(), 0, QModelIndex())
        points_folder_index = self.layer_model.index(selection_layer.points_folder.child_number(),
                                                     0, selection_index)
        if self.layer_model.rowCount(points_folder_index) > 0:
            self.layer_model.request_expand.emit(selection_index)
            self.layer_model.request_expand.emit(points_folder_index)
        profiles_folder_index = self.layer_model.index(
            selection_layer.profiles_folder.child_number(), 0, selection_index)
        if self.layer_model.rowCount(profiles_folder_index) > 0:
            self.layer_model.request_expand.emit(selection_index)
            self.layer_model.request_expand.emit(profiles_folder_index)
        references_folder_index = self.layer_model.index(
            selection_layer.references_folder.child_number(), 0, selection_index)
        if self.layer_model.rowCount(references_folder_index) > 0:
            self.layer_model.request_expand.emit(selection_index)
            self.layer_model.request_expand.emit(references_folder_index)
        self.flip_h = input_dict.get("flip_h", False)
        self.flip_v = input_dict.get("flip_v", False)
        self.current_band_index = input_dict.get("current_band", self.current_band_index)
        self.request_colormap.emit(input_dict.get("colormap", "greyscale"))
        if "total_histogram" in input_dict and "band_histograms" in input_dict:
            if not self.loader.set_histograms(input_dict["total_histogram"],
                                              input_dict["band_histograms"]):
                logger.warning("Histograms exist but are invalid")
                self.request_histograms.emit(None)
        else:
            logger.info("Histograms are missing, computing them")
            self.request_histograms.emit(None)
        self.colormap_v0 = input_dict.get("colormap_v0", self.colormap_v0)
        self.colormap_v1 = input_dict.get("colormap_v1", self.colormap_v1)
        self.v0_v1_changed.emit(self.colormap_v0, self.colormap_v1)
        if self.current_band_index not in self.loader.dataset.indexes:
            logger.warning(
                f"current_band {self.current_band_index} does not match any existing band, change to 1")
            self.current_band_index = 1
        return True
